
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>

#include <SDL/SDL.h>
#include <SDL/SDL_video.h>
#include <SDL/SDL_keysym.h>
#include <GL/gl.h>

#include "defines.h"
#include "main.h"
#include "gui.h"
#include "ogl.h"
#include "../client/oload.h"
#include "../client/action.h"
#include "../client/network.h"
#include "../client/connection.h"
#include "../client/collision.h"
#include "../client/console.h"


extern volatile int connection_status;
extern struct game_object player_object;
extern struct vector gravity, new_velocity_vector;
extern double collision_txt_speed;
extern char server_name[256];
extern int actual_socket, actual_port;
extern int player_motion, status, player_status, collision_occurred;
extern int show_blocks, status, texturing_status;

int console_status = OFF, cross_status = OFF, lod_status = ON, lantern_status = ON;
double player_x = 0, player_y = 0, position_x = 0, position_y = 0, position_z = 0;

const SDL_VideoInfo *sdl_video_info = NULL;

struct vector jump_normal;
double jump_table[] = {
  0.8, 0.8, 0.7, 0.7, 0.7,
  0.6, 0.6, 0.5, 0.4, 0.3,
};
int jump_counter = 0;


int init_window_main(void) {

  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER) < 0) {
    fprintf(stderr, "INIT_WINDOW_MAIN: Video initialization failed: %s\n", SDL_GetError());
    return FAILED;
  }

  sdl_video_info = SDL_GetVideoInfo();

  SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

  if (SDL_SetVideoMode(SCREEN_X, SCREEN_Y, sdl_video_info->vfmt->BitsPerPixel, SDL_OPENGL) == 0) {
    fprintf(stderr, "INIT_WINDOW_MAIN: Video mode initialization failed: %s\n", SDL_GetError());
    return FAILED;
  }

  SDL_EnableUNICODE(1);

  return SUCCEEDED;
}


int main_window_set_title(void) {

  char b[256];


  if (connection_status == OFF) {
    sprintf(b, "%s @ Not connected", player_object.name);
  }
  else {
    sprintf(b, "%s @ %s:%d", player_object.name, server_name, actual_port);
  }

  SDL_WM_SetCaption(b, b);

  return SUCCEEDED;
}


int keyboard_io(void) {

  SDL_Event event;
  SDLMod mod;
  int i;


  while (SDL_PollEvent(&event)) {
    switch (event.type) {

    /* mouse */
    case SDL_MOUSEBUTTONDOWN:
      /* pick up */
      if (event.button.button == 3)
	action_pick_up();
      break;

    /* key up */
    case SDL_KEYUP:
      if (console_status == OFF) {
	switch (event.key.keysym.sym) {
	case SDLK_LEFT:
	case SDLK_a:
	  player_motion &= ~PLAYER_MOTION_LEFT;
	  break;
	case SDLK_RIGHT:
	case SDLK_d:
	  player_motion &= ~PLAYER_MOTION_RIGHT;
	  break;
	case SDLK_UP:
	case SDLK_w:
	  player_motion &= ~PLAYER_MOTION_UP;
	  break;
	case SDLK_DOWN:
	case SDLK_s:
	  player_motion &= ~PLAYER_MOTION_DOWN;
	  break;
	default:
	  break;
	}
      }
      break;

    /* key down */
    case SDL_KEYDOWN:
      i = event.key.keysym.sym;
      if (console_status == ON) {
	if ((i >= SDLK_a && i <= SDLK_z) || (i >= SDLK_0 && i <= SDLK_9))
	  console_type_char(event.key.keysym.sym);
	else if (i == SDLK_BACKSPACE)
	  console_erase_char();
	else if (i == SDLK_RETURN)
	  console_type_char(0);
	else if (i == SDLK_TAB || i == SDLK_ESCAPE)
	  console_status = OFF;
	else if (i >= SDLK_SPACE && i <= SDLK_SLASH)
	  console_type_char(event.key.keysym.sym);
	else if (i == SDLK_F8)
	  position_save();
	else if (i == SDLK_F9)
	  position_jump();
	else if (i == SDLK_HOME)
	  toggle_lod();
      }
      else {
	switch (event.key.keysym.sym) {
	case SDLK_i:
	  SDL_WM_IconifyWindow();
	  break;
	case SDLK_l:
	  if (lantern_status == ON)
	    lantern_status = OFF;
	  else
	    lantern_status = ON;
	  send_lantern();
	  break;
	case SDLK_t:
	  if (texturing_status == ON)
	    texturing_status = OFF;
	  else
	    texturing_status = ON;
	  break;
	case SDLK_F8:
	  position_save();
	  break;
	case SDLK_F9:
	  position_jump();
	  break;
	case SDLK_HOME:
	  toggle_lod();
	  break;
	case SDLK_TAB:
	  console_status = ON;
	  cross_status = OFF;
	  break;
	case SDLK_PLUS:
	  show_blocks++;
	  send_visibility();
	  break;
	case SDLK_MINUS:
	  if (show_blocks > 2) {
	    show_blocks--;
	    send_visibility();
	  }
	  break;
	case SDLK_e:
	  if (cross_status == OFF)
	    cross_status = ON;
	  else
	    cross_status = OFF;
	  break;
	case SDLK_LEFT:
	case SDLK_a:
	  player_motion |= PLAYER_MOTION_LEFT;
	  break;
	case SDLK_RIGHT:
	case SDLK_d:
	  player_motion |= PLAYER_MOTION_RIGHT;
	  break;
	case SDLK_UP:
	case SDLK_w:
	  player_motion |= PLAYER_MOTION_UP;
	  break;
	case SDLK_DOWN:
	case SDLK_s:
	  player_motion |= PLAYER_MOTION_DOWN;
	  break;
	case SDLK_SPACE:
	  if (jump_counter == 0) {
	    jump_counter = 9;
	  }
	  break;
	case SDLK_ESCAPE:
	  send_quit();
	  status |= STATUS_QUIT;
	  break;
	default:
	  break;
	}
	break;
      }
    }
  }

  mod = SDL_GetModState();
  if (((mod & KMOD_LSHIFT) != 0) || ((mod & KMOD_RSHIFT) != 0))
    player_motion |= PLAYER_MOTION_RUN;
  else
    player_motion &= ~PLAYER_MOTION_RUN;

  return SUCCEEDED;
}


int sdl_filter_events(const SDL_Event *event) {

  if (event->type == SDL_QUIT)
    return 1;

  if (event->type == SDL_MOUSEMOTION) {
    player_x += event->motion.xrel / (double)SCREEN_X * 360;
    player_y += event->motion.yrel / (double)SCREEN_Y * 360;
    if (player_x < 0)
      player_x += 360;
    if (player_x >= 360)
      player_x -= 360;
    if (player_y < -70)
      player_y = -70;
    if (player_y > 70)
      player_y = 70;
    player_status |= PLAYER_STATUS_MOVED;
  }

  return 1;
}


int compute_new_position(void) {

  struct vector pos, move;
  double p, speed, x, y, z;
  int motion;


  /* start with gravity applied */
  move.x = gravity.x + player_object.vx;
  move.y = gravity.y + player_object.vy;
  move.z = gravity.z + player_object.vz;
  motion = OFF;

  if (player_motion & PLAYER_MOTION_RUN)
    speed = 1.8 * collision_txt_speed;
  else
    speed = 1.3 * collision_txt_speed;

  if (player_object.status & GO_COLLIDED) {
    if (player_motion & PLAYER_MOTION_UP) {
      p = 2 * M_PI * player_x / 360;
      move.x += cos(p) * speed;
      move.z += sin(p) * speed;
      motion = ON;
    }
    if (player_motion & PLAYER_MOTION_DOWN) {
      p = 2 * M_PI * player_x / 360;
      move.x -= cos(p) * speed / 2;
      move.z -= sin(p) * speed / 2;
      motion = ON;
    }
    if (player_motion & PLAYER_MOTION_LEFT) {
      p = player_x + 90;
      if (p >= 360)
	p -= 360;
      p = 2 * M_PI * p / 360;
      move.x -= cos(p) * speed / 1.5;
      move.z -= sin(p) * speed / 1.5;
      motion = ON;
    }
    if (player_motion & PLAYER_MOTION_RIGHT) {
      p = player_x - 90;
      if (p < 0)
	p += 360;
      p = 2 * M_PI * p / 360;
      move.x -= cos(p) * speed / 1.5;
      move.z -= sin(p) * speed / 1.5;
      motion = ON;
    }
  }

  if (jump_counter != 0) {
    p = jump_table[jump_counter];
    move.x += p * jump_normal.x;
    move.y += p * jump_normal.y;
    move.z += p * jump_normal.z;
    jump_counter--;
  }

  pos.x = player_object.x;
  pos.y = player_object.y;
  pos.z = player_object.z;
  collision_detection(&pos, &move);

  /* update collided status */
  if (collision_occurred == YES)
    player_object.status |= GO_COLLIDED;
  else
    player_object.status &= ~GO_COLLIDED;

  player_object.slide_plane_vector.x = new_velocity_vector.x;
  player_object.slide_plane_vector.y = new_velocity_vector.y;
  player_object.slide_plane_vector.z = new_velocity_vector.z;

  x = pos.x - player_object.x;
  y = pos.y - player_object.y;
  z = pos.z - player_object.z;
  p = sqrt(x*x + y*y + z*z);

  /* passed collision checks? */
  if (p > EPSILON/100) {
    player_object.x = pos.x;
    player_object.y = pos.y;
    player_object.z = pos.z;
    player_status |= PLAYER_STATUS_MOVED;

    player_object.vx = player_object.slide_plane_vector.x * 0.5;
    player_object.vy = player_object.slide_plane_vector.y * 0.5;
    player_object.vz = player_object.slide_plane_vector.z * 0.5;
  }
  else {
    player_object.vx = 0;
    player_object.vy = 0;
    player_object.vz = 0;
  }

  return SUCCEEDED;
}


/*  \3/
    2 0
    /1\  */

int get_player_x_segment(void) {

  if (player_x > 315 || player_x < 45)
    return 0;
  if (player_x >= 45 && player_x < 135)
    return 1;
  if (player_x >= 135 && player_x < 225)
    return 2;
  return 3;
}


void position_save(void) {

  position_x = player_object.x;
  position_y = player_object.y;
  position_z = player_object.z;
}


void position_jump(void) {

  player_object.x = position_x;
  player_object.y = position_y;
  player_object.z = position_z;
}
