
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <SDL/SDL_video.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "oload.h"
#include "txt.h"
#include "ogl.h"
#include "game_object.h"
#include "../common/mesh2.h"


extern struct object_object **world_map;
struct object_object **o_lod_1 = NULL, **o_lod_2 = NULL, **o_lod_2_tmp;
extern int world_x, world_y, world_z, lod_status;

char world_name[256];
char *object_occluder = NULL;
int collision_status;
int object_max_id;
int *object_type = NULL;


int object_load_list(void) {

  char l1[64], l2[64], c[256];
  int id, type, occluder;
  FILE *f;


  f = fopen("../data/obj/list.txt", "rb");
  if (f == NULL) {
    fprintf(stderr, "OBJECT_LOAD_LIST: Could not open file \"../data/obj/list.txt\".\n");
    return FAILED;
  }

  object_max_id = 0;

  /* get pass the comment section */
  fscanf(f, "%s", c);
  while (strcmp(c, "GO!") != 0)
    if (fscanf(f, "%s", c) == EOF)
      break;

  while (fscanf(f, "%d %*d %*d %*s %*s", &id) != EOF) {
    if (object_max_id < id)
      object_max_id = id;
  }

  fseek(f, 0, SEEK_SET);

  /* allocate array of pointers (to objects) */
  o_lod_1 = malloc(sizeof(struct object_object *) * (object_max_id+1));
  o_lod_2 = malloc(sizeof(struct object_object *) * (object_max_id+1));
  object_type = malloc(sizeof(int) * (object_max_id+1));
  object_occluder = malloc(object_max_id+1);

  if (o_lod_1 == NULL || o_lod_2 == NULL || object_type == NULL || object_occluder == NULL) {
    fprintf(stderr, "OBJECT_LOAD_LIST: Out of memory error.\n");
    return FAILED;
  }

  for (id = 0; id < object_max_id+1; id++) {
    *(o_lod_1 + id) = NULL;
    *(o_lod_2 + id) = NULL;
  }

  /* get past the comment section */
  fscanf(f, "%s", c);
  while (strcmp(c, "GO!") != 0)
    if (fscanf(f, "%s", c) == EOF)
      break;

  while (fscanf(f, "%d %d %d %63s %63s", &id, &type, &occluder, l1, l2) != EOF) {
    /* lod 1 */
    strcpy(c, "../data/obj/");
    strcat(c, l1);
    if (object_load(c, o_lod_1 + id, YES) == FAILED)
      return FAILED;

    /* lod 2 */
    strcpy(c, "../data/obj/");
    strcat(c, l2);
    if (object_load(c, o_lod_2 + id, YES) == FAILED)
      return FAILED;

    object_type[id] = type;
    object_occluder[id] = occluder;
  }

  fclose(f);

  return SUCCEEDED;
}


void object_get(int id, int lod, struct object_object **o) {

  if (lod == 1)
    *o = *(o_lod_1 + id);
  else
    *o = *(o_lod_2 + id);

  if (*o == NULL)
    fprintf(stderr, "GET_OBJECT: Could not find object %d.\n", id);
}


int object_load(char *name, struct object_object **v, int display_load_errors) {

  struct draw_item *di;
  struct mesh2 *m;
  float *ve;
  int i;


  if (mesh2_read_binary(name, &m, display_load_errors) == FAILED) {
    *v = NULL;
    return FAILED;
  }

  *v = malloc(sizeof(struct object_object));
  if (*v == NULL) {
    fprintf(stderr, "OBJECT_LOAD: Out of memory.\n");
    mesh2_free(m);
    return FAILED;
  }

  /* compute bounding sphere r for the whole object */
  mesh2_compute_bounding_sphere(m, (*v)->sdi.bounding_sphere);

  m->collision_status = ON;

  /* estimate triangles with spheres */
  mesh2_compute_triangle_bounding_spheres(m, YES);

  (*v)->mesh = m;
  (*v)->style = 0;

  di = &((*v)->sdi);

  /* init sdi */
  di->type = DI_TYPE_OO;
  di->visibility_counter = 0;
  di->visibility_status = VIS_IS_HIDDEN;
  di->data = (void *)*v;

  /* compute the aabb */
  ve = m->vertex_list;
  di->aabb[AABB_MINX] = ve[0];
  di->aabb[AABB_MAXX] = ve[0];
  di->aabb[AABB_MINY] = ve[1];
  di->aabb[AABB_MAXY] = ve[1];
  di->aabb[AABB_MINZ] = ve[2];
  di->aabb[AABB_MAXZ] = ve[2];
  ve += 3;

  for (i = 1; i < m->vertices; i++, ve += 3) {
    if (ve[0] < di->aabb[AABB_MINX])
      di->aabb[AABB_MINX] = ve[0];
    if (ve[0] > di->aabb[AABB_MAXX])
      di->aabb[AABB_MAXX] = ve[0];
    if (ve[1] < di->aabb[AABB_MINY])
      di->aabb[AABB_MINY] = ve[1];
    if (ve[1] > di->aabb[AABB_MAXY])
      di->aabb[AABB_MAXY] = ve[1];
    if (ve[2] < di->aabb[AABB_MINZ])
      di->aabb[AABB_MINZ] = ve[2];
    if (ve[2] > di->aabb[AABB_MAXZ])
      di->aabb[AABB_MAXZ] = ve[2];
  }

  return SUCCEEDED;
}


int toggle_lod(void) {

  if (lod_status == ON) {
    lod_status = OFF;
    o_lod_2_tmp = o_lod_2;
    o_lod_2 = o_lod_1;
  }
  else {
    lod_status = ON;
    o_lod_2 = o_lod_2_tmp;
  }

  return SUCCEEDED;
}
