
/* (c) 2002-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <SDL/SDL_video.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "oload.h"
#include "txt.h"
#include "ogl.h"
#include "world.h"
#include "../common/mesh2.h"



extern struct game_object player_object;
extern int show_blocks, frame_time;

struct object_object **world_map;
char *world_map_status;
char world_name[256];
int world_x, world_y, world_z;
int cubes_loaded = 0, cubes_max = 4*4*4*16;
int cubes_load_all = YES;



int world_init(char *n) {

  FILE *f;
  char c[256];
  int i;


  strcpy(c, "../data/wld/");
  strcat(c, n);
  strcat(c, "/");
  strcpy(world_name, c);
  strcat(c, "info.txt");
  f = fopen(c, "rb");
  if (f == NULL) {
    fprintf(stderr, "WORLD_INIT: Could not open \"%s\".\n", c);
    return FAILED;
  }

  fscanf(f, "%d %d %d", &world_x, &world_y, &world_z);
  fclose(f);

  world_map = malloc(sizeof(struct object_object *) * world_x * world_y * world_z);
  if (world_map == NULL) {
    fprintf(stderr, "WORLD_INIT: Out of memory.\n");
    return FAILED;
  }

  world_map_status = malloc(world_x * world_y * world_z);
  if (world_map_status == NULL) {
    fprintf(stderr, "WORLD_INIT: Out of memory.\n");
    free(world_map);
    world_map = NULL;
    return FAILED;
  }

  for (i = 0; i < world_x*world_y*world_z; i++) {
    world_map[i] = NULL;
    world_map_status[i] = MAP_NOT_LOADED;
  }

  /* load all the missing cubes at startup */
  cubes_load_all = YES;

  return SUCCEEDED;
}


int _world_rb_cmp(const void *a, const void *b) {

  int *x, *y;


  x = (int *)a;
  y = (int *)b;

  if (*(x+3) == *(y+3))
    return 0;
  if (*(x+3) > *(y+3))
    return -1;

  return 1;
}


int _world_free_object(int i) {

  struct object_object *t;


  world_map_status[i] = MAP_NOT_LOADED;
  t = world_map[i];
  world_map[i] = NULL;

  mesh2_free(t->mesh);

  free(t);

  return SUCCEEDED;
}


int world_dynamic_refresh(void) {

  struct object_object *v;
  double dx, dy, dz;
  char name[256];
  static int rb_size = 0, *rb = NULL, rb_write = 0;
  int ax, ay, az, ex, ey, ez, x, y, z, i, p, loaded;


  if (rb == NULL) {
    rb = malloc(sizeof(int)*cubes_max*4);
    if (rb == NULL) {
      fprintf(stderr, "WORLD_DYNAMIC_REFRESH: Out of memory error.\n");
      return FAILED;
    }
    rb_size = cubes_max;
  }

  ax = (player_object.x / CUBE_SIDE) - show_blocks - 1;
  ay = (player_object.y / CUBE_SIDE) - show_blocks - 1;
  az = (player_object.z / CUBE_SIDE) - show_blocks - 1;
  ex = ax + 2*show_blocks + 1 + 2;
  ey = ay + 2*show_blocks + 1 + 2;
  ez = az + 2*show_blocks + 1 + 2;

  if (ax < 0)
    ax = 0;
  if (ay < 0)
    ay = 0;
  if (az < 0)
    az = 0;
  if (ex > world_x)
    ex = world_x;
  if (ey > world_y)
    ey = world_y;
  if (ez > world_z)
    ez = world_z;

  loaded = 0;

  for (z = az; z < ez; z++) {
    for (y = ay; y < ey; y++) {
      for (x = ax; x < ex; x++) {

	p = z*world_y*world_x + y*world_x + x;
	if (world_map_status[p] == MAP_NOT_LOADED) {
	  sprintf(name, "%s%.3d-%.3d-%.3d.me2", world_name, z, y, x);
	  if (object_load(name, &v, NO) == FAILED) {
	    world_map[p] = NULL;
	    world_map_status[p] = MAP_LOADED;
	  }
	  else {
	    world_map[p] = v;
	    world_map_status[p] = MAP_LOADED;

	    /* free most distant cubes */
	    if (cubes_loaded == rb_size) {

	      for (i = 0; i < rb_size; i++) {
		dx = (player_object.x/CUBE_SIDE) - rb[i*4+0];
		dy = (player_object.y/CUBE_SIDE) - rb[i*4+1];
		dz = (player_object.z/CUBE_SIDE) - rb[i*4+2];
		rb[i*4+3] = dx*dx + dy*dy + dz*dz;
	      }

	      qsort(rb, rb_size, 4*sizeof(int), _world_rb_cmp);

	      /* free 10% of the cache */
	      for (i = 0; i < rb_size/10; i++)
		_world_free_object(rb[i*4+2]*world_y*world_x + rb[i*4+1]*world_x + rb[i*4+0]);

	      cubes_loaded -= rb_size/10;
	      rb_write = 0;
	    }

	    rb[rb_write*4+0] = x;
	    rb[rb_write*4+1] = y;
	    rb[rb_write*4+2] = z;
	    rb_write = (rb_write+1)%rb_size;
	    cubes_loaded++;
	    loaded++;

	    /* abort if enough cubes have been loaded this round */
	    if (cubes_load_all == NO && loaded == 1) {
	      z = ez;
	      y = ey;
	      x = ex;
	      continue;
	    }
	  }
	}
      }
    }
  }

  /* loading of all cubes must be enabled explicitly each time */
  cubes_load_all = NO;

  return SUCCEEDED;
}
