
/* (c) 2002-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "frustum.h"


/* this code is based on mark morley's code, which can be found at
   http://www.markmorley.com/opengl/frustumculling.html */

/* put the current opengl frustum into f */
void frustum_get(float *f) {

  float p[16], m[16], c[16], t;


  glGetFloatv(GL_PROJECTION_MATRIX, p);
  glGetFloatv(GL_MODELVIEW_MATRIX, m);

  /* combine the two matrices (multiply projection by modelview) */
  c[0] = m[0] * p[0] + m[1] * p[4] + m[2] * p[8] + m[3] * p[12];
  c[1] = m[0] * p[1] + m[1] * p[5] + m[2] * p[9] + m[3] * p[13];
  c[2] = m[0] * p[2] + m[1] * p[6] + m[2] * p[10] + m[3] * p[14];
  c[3] = m[0] * p[3] + m[1] * p[7] + m[2] * p[11] + m[3] * p[15];

  c[4] = m[4] * p[0] + m[5] * p[4] + m[6] * p[8] + m[7] * p[12];
  c[5] = m[4] * p[1] + m[5] * p[5] + m[6] * p[9] + m[7] * p[13];
  c[6] = m[4] * p[2] + m[5] * p[6] + m[6] * p[10] + m[7] * p[14];
  c[7] = m[4] * p[3] + m[5] * p[7] + m[6] * p[11] + m[7] * p[15];

  c[8] = m[8] * p[0] + m[9] * p[4] + m[10] * p[8] + m[11] * p[12];
  c[9] = m[8] * p[1] + m[9] * p[5] + m[10] * p[9] + m[11] * p[13];
  c[10] = m[8] * p[2] + m[9] * p[6] + m[10] * p[10] + m[11] * p[14];
  c[11] = m[8] * p[3] + m[9] * p[7] + m[10] * p[11] + m[11] * p[15];

  c[12] = m[12] * p[0] + m[13] * p[4] + m[14] * p[8] + m[15] * p[12];
  c[13] = m[12] * p[1] + m[13] * p[5] + m[14] * p[9] + m[15] * p[13];
  c[14] = m[12] * p[2] + m[13] * p[6] + m[14] * p[10] + m[15] * p[14];
  c[15] = m[12] * p[3] + m[13] * p[7] + m[14] * p[11] + m[15] * p[15];

  /* extract the numbers for the RIGHT plane */
  f[0*4 + 0] = c[3] - c[0];
  f[0*4 + 1] = c[7] - c[4];
  f[0*4 + 2] = c[11] - c[8];
  f[0*4 + 3] = c[15] - c[12];

  /* extract the numbers for the LEFT plane */
  f[1*4 + 0] = c[3] + c[0];
  f[1*4 + 1] = c[7] + c[4];
  f[1*4 + 2] = c[11] + c[8];
  f[1*4 + 3] = c[15] + c[12];

  /* extract the BOTTOM plane */
  f[2*4 + 0] = c[3] + c[1];
  f[2*4 + 1] = c[7] + c[5];
  f[2*4 + 2] = c[11] + c[9];
  f[2*4 + 3] = c[15] + c[13];

  /* extract the TOP plane */
  f[3*4 + 0] = c[3] - c[1];
  f[3*4 + 1] = c[7] - c[5];
  f[3*4 + 2] = c[11] - c[9];
  f[3*4 + 3] = c[15] - c[13];

  /* extract the FAR plane */
  f[4*4 + 0] = c[3] - c[2];
  f[4*4 + 1] = c[7] - c[6];
  f[4*4 + 2] = c[11] - c[10];
  f[4*4 + 3] = c[15] - c[14];

  /* extract the NEAR plane */
  f[5*4 + 0] = c[3] + c[2];
  f[5*4 + 1] = c[7] + c[6];
  f[5*4 + 2] = c[11] + c[10];
  f[5*4 + 3] = c[15] + c[14];

  /* normalize the results */
  t = sqrt(f[0*4 + 0] * f[0*4 + 0] + f[0*4 + 1] * f[0*4 + 1] + f[0*4 + 2] * f[0*4 + 2]);
  f[0*4 + 0] /= t;
  f[0*4 + 1] /= t;
  f[0*4 + 2] /= t;
  f[0*4 + 3] /= t;
  t = sqrt(f[1*4 + 0] * f[1*4 + 0] + f[1*4 + 1] * f[1*4 + 1] + f[1*4 + 2] * f[1*4 + 2]);
  f[1*4 + 0] /= t;
  f[1*4 + 1] /= t;
  f[1*4 + 2] /= t;
  f[1*4 + 3] /= t;
  t = sqrt(f[2*4 + 0] * f[2*4 + 0] + f[2*4 + 1] * f[2*4 + 1] + f[2*4 + 2] * f[2*4 + 2]);
  f[2*4 + 0] /= t;
  f[2*4 + 1] /= t;
  f[2*4 + 2] /= t;
  f[2*4 + 3] /= t;
  t = sqrt(f[3*4 + 0] * f[3*4 + 0] + f[3*4 + 1] * f[3*4 + 1] + f[3*4 + 2] * f[3*4 + 2]);
  f[3*4 + 0] /= t;
  f[3*4 + 1] /= t;
  f[3*4 + 2] /= t;
  f[3*4 + 3] /= t;
  t = sqrt(f[4*4 + 0] * f[4*4 + 0] + f[4*4 + 1] * f[4*4 + 1] + f[4*4 + 2] * f[4*4 + 2]);
  f[4*4 + 0] /= t;
  f[4*4 + 1] /= t;
  f[4*4 + 2] /= t;
  f[4*4 + 3] /= t;
  t = sqrt(f[5*4 + 0] * f[5*4 + 0] + f[5*4 + 1] * f[5*4 + 1] + f[5*4 + 2] * f[5*4 + 2]);
  f[5*4 + 0] /= t;
  f[5*4 + 1] /= t;
  f[5*4 + 2] /= t;
  f[5*4 + 3] /= t;
}


/* test if the sphere (x, y, z, r) is inside the frustum f */
int frustum_sphere_inside(float *f, float x, float y, float z, float r) {

  r = -r;
  if (f[0]*x + f[1]*y + f[2]*z + f[3] <= r)
    return NO;
  if (f[4]*x + f[5]*y + f[6]*z + f[7] <= r)
    return NO;
  if (f[8]*x + f[9]*y + f[10]*z + f[11] <= r)
    return NO;
  if (f[12]*x + f[13]*y + f[14]*z + f[15] <= r)
    return NO;
  if (f[16]*x + f[17]*y + f[18]*z + f[19] <= r)
    return NO;
  if (f[20]*x + f[21]*y + f[22]*z + f[23] <= r)
    return NO;

  return YES;
}
