
/* (c) 2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <GL/gl.h>
#include <GL/glu.h>



void quaternion_rotation_matrix(float ax, float ay, float az, GLfloat *rm) {

  float qx[2], qy[3], qz[4], t[4], q[4], wz, xy, xx, yy, zz, wx, wy, xz, yz;


  /* create one quaternion for each axis */
  ax /= 2;
  ay /= 2;
  az /= 2;

  qx[0] = cos(ax);
  qx[1] = sin(ax);
  qy[0] = cos(ay);
  qy[2] = sin(ay);
  qz[0] = cos(az);
  qz[3] = sin(az);

  /* now get t=qx*qy */
  t[0] = qx[0]*qy[0];
  t[1] = qx[1]*qy[0];
  t[2] = qx[0]*qy[2];
  t[3] = qx[1]*qy[2];

  /* next get q=t*qz */
  q[0] = t[0]*qz[0] - t[3]*qz[3];
  q[1] = t[1]*qz[0] + t[2]*qz[3];
  q[2] = t[2]*qz[0] - t[1]*qz[3];
  q[3] = t[0]*qz[3] + t[3]*qz[0];

  /* quaternion -> matrix */
  xx = 2*q[1]*q[1];
  yy = 2*q[2]*q[2];
  zz = 2*q[3]*q[3];
  wx = 2*q[0]*q[1];
  wz = 2*q[0]*q[3];
  xy = 2*q[1]*q[2];
  wy = 2*q[0]*q[2];
  xz = 2*q[1]*q[3];
  yz = 2*q[2]*q[3];

  rm[0] = 1 - yy - zz;
  rm[4] = xy - wz;
  rm[8] = xz + wy;
  rm[12] = 0;

  rm[1] = xy + wz;
  rm[5] = 1 - xx - zz;
  rm[9] = yz - wx;
  rm[13] = 0;

  rm[2] = xz - wy;
  rm[6] = yz + wx;
  rm[10] = 1 - xx - yy;
  rm[14] = 0;

  rm[3] = 0;
  rm[7] = 0;
  rm[11] = 0;
  rm[15] = 1;
}


/*
  q1[0] = qx[0]*qy[0] - qx[1]*qy[1] - qx[2]*qy[2] - qx[3]*qy[3];
  q1[1] = qx[0]*qy[1] + qx[1]*qy[0] + qx[2]*qy[3] - qx[3]*qy[2];
  q1[2] = qx[0]*qy[2] + qx[2]*qy[0] + qx[3]*qy[1] - qx[1]*qy[3];
  q1[3] = qx[0]*qy[3] + qx[3]*qy[0] + qx[1]*qy[2] - qx[2]*qy[1];
*/
