
/* (c) 2002-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "random.h"
#include "defines.h"



/* a gaussian (0,l) random generator (uses box-muller transformation) */
double random_gaussian(double l) {

  double r, x1, x2;
  static double s;
  static int x = 0;


  /* is there one already in memory? */
  if (x == 1) {
    x = 0;
    return s;
  }

  do {
    x1 = 2.0*(rand()/(double)RAND_MAX) - 1.0;
    x2 = 2.0*(rand()/(double)RAND_MAX) - 1.0;
    r = x1*x1 + x2*x2;
  } while (r >= 1.0);

  r = sqrt(-2.0*log(r)/r);
  r = r*l/2.0;

  x = 1;
  s = x2*r;

  return x1*r;
}


static int _is_point_inside_a_triangle(double *p, double *t) {

  double u[3], v[3], w[3], uu, uv, vv, wu, wv, s, n, d;


  /* u ~ 0->1 */
  u[0] = t[3*1 + 0] - t[3*0 + 0];
  u[1] = t[3*1 + 1] - t[3*0 + 1];
  u[2] = t[3*1 + 2] - t[3*0 + 2];

  /* v ~ 0->2 */
  v[0] = t[3*2 + 0] - t[3*0 + 0];
  v[1] = t[3*2 + 1] - t[3*0 + 1];
  v[2] = t[3*2 + 2] - t[3*0 + 2];

  /* w ~ 0->p */
  w[0] = p[0] - t[3*0 + 0];
  w[1] = p[1] - t[3*0 + 1];
  w[2] = p[2] - t[3*0 + 2];

  uu = u[0]*u[0] + u[1]*u[1] + u[2]*u[2];
  uv = u[0]*v[0] + u[1]*v[1] + u[2]*v[2];
  vv = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
  wu = w[0]*u[0] + w[1]*u[1] + w[2]*u[2];
  wv = w[0]*v[0] + w[1]*v[1] + w[2]*v[2];

  d = uv*uv - uu*vv;

  s = (uv*wv - vv*wu)/d;
  if (s < 0.0 || s > 1.0)
    return NO;

  n = (uv*wu - uu*wv)/d;
  if (n < 0.0 || (s+n) > 1.0)
    return NO;

  return YES;
}


int random_point_inside_a_triangle(double *tri, double *x, double *y, double *z) {

  double a, d, dx, dy, dz, ax, ay, az;
  double p[3];


  if (tri == NULL)
    return FAILED;

  while (1) {
    d = rand()/(double)RAND_MAX;
    a = rand()/(double)RAND_MAX;

    /* scale the edge a->c by a random value [0,1] */
    dx = tri[2*3 + 0] - tri[1*3 + 0];
    dy = tri[2*3 + 1] - tri[1*3 + 1];
    dz = tri[2*3 + 2] - tri[1*3 + 2];
    dx *= d;
    dy *= d;
    dz *= d;

    /* scale the edge a->b by a random value [0,1] */
    ax = tri[0*3 + 0] - tri[1*3 + 0];
    ay = tri[0*3 + 1] - tri[1*3 + 1];
    az = tri[0*3 + 2] - tri[1*3 + 2];
    ax *= a;
    ay *= a;
    az *= a;

    p[0] = tri[1*3 + 0]+ dx + ax;
    p[1] = tri[1*3 + 1]+ dy + ay;
    p[2] = tri[1*3 + 2]+ dz + az;

    if (_is_point_inside_a_triangle(p, tri) == YES)
      break;
  }

  *x = p[0];
  *y = p[1];
  *z = p[2];

  return SUCCEEDED;
}
