
/* (c) 2002-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "defines.h"
#include "oload.h"


struct object_object *objects_first = NULL, *objects_last = NULL;
struct object_mesh *meshes_first = NULL, *meshes_last = NULL;
int meshes_total;


int object_load_ac3d(char *name) {

  struct object_mesh *mesh = NULL;
  double *d;
  float loc_x, loc_y, loc_z, x, y, z, *e;
  FILE *f;
  char tmp[256], *c;
  int objects_count = 0, i, *g, p, r, skipped = 0;


  f = fopen(name, "rb");
  if (f == NULL) {
    fprintf(stderr, "OBJECT_LOAD_AC3D: Could not open \"%s\".\n", name);
    return FAILED;
  }

  /* check format */
  fscanf(f, "%s", tmp);
  if (strcmp(tmp, "AC3Db") != 0) {
    fprintf(stderr, "OBJECT_LOAD_AC3D: File \"%s\" is of unknown format \"%s\".\n", name, tmp);
    fclose(f);
    return FAILED;
  }

  /* skip materials */
  fscanf(f, "%s", tmp);
  while (strcmp(tmp, "MATERIAL") == 0) {
    /* name */
    fscanf(f, "%s", tmp);
    /* rgb */
    fscanf(f, "%*s %*f %*f %*f");
    /* amb */
    fscanf(f, "%*s %*f %*f %*f");
    /* emis */
    fscanf(f, "%*s %*f %*f %*f");
    /* spec */
    fscanf(f, "%*s %*f %*f %*f");
    /* shi */
    fscanf(f, "%*s %*d");
    /* trans */
    fscanf(f, "%*s %*f");
    fscanf(f, "%s", tmp);
  }

  /* load objects */
  while (strcmp(tmp, "OBJECT") == 0) {
    /* type (world, poly, group */
    fscanf(f, "%s", tmp);
    /* world */
    if (strcmp(tmp, "world") == 0) {
      fscanf(f, "%*s %d", &meshes_total);
      objects_count = meshes_total;
      fscanf(f, "%s", tmp);
      continue;
    }
    /* group */
    if (strcmp(tmp, "group") == 0) {
      fprintf(stderr, "OBJECT_LOAD_AC3D: Unknown object type \"group\".\n");
      fclose(f);
      return FAILED;
    }
    /* poly */
    if (strcmp(tmp, "poly") == 0) {
      objects_count--;
      if (objects_count < 0) {
	fprintf(stderr, "OBJECT_LOAD_AC3D: Error with object count.\n");
	fclose(f);
	return FAILED;
      }

      /* allocate a new mesh */
      mesh = malloc(sizeof(struct object_mesh));
      if (mesh == NULL) {
	fprintf(stderr, "OBJECT_LOAD_AC3D: Out of memory error.\n");
	fclose(f);
	return FAILED;
      }

      /* poly name */
      fscanf(f, "%s", tmp);
      if (strcmp(tmp, "name") != 0) {
	fprintf(stderr, "OBJECT_LOAD_AC3D: No poly name!\n");
	fclose(f);
	return FAILED;
      }
      fscanf(f, "%s", mesh->name);

      /* loc */
      loc_x = 0;
      loc_y = 0;
      loc_z = 0;

      fscanf(f, "%s", tmp);
      if (strcmp(tmp, "loc") == 0) {
	fscanf(f, "%f %f %f", &loc_x, &loc_y, &loc_z);
	fscanf(f, "%s", tmp);
      }

      mesh->textures = (int *)malloc(256);
      if (mesh->textures == NULL) {
	fprintf(stderr, "OBJECT_LOAD_AC3D: Out of memory error.\n");
	fclose(f);
	return FAILED;
      }

      /* texture */
      if (strcmp(tmp, "texture") == 0) {
	fscanf(f, "%s", tmp);

	/* take only the final file name */
	for (i = strlen(tmp) - 1; i >= 0 && tmp[i] != '/'; i--);
	if (i == -1)
	  i = 0;
	for (p = 0, i++; tmp[i] != '"'; i++, p++)
	  *(((char *)mesh->textures) + p) = tmp[i];
	*(((char *)mesh->textures) + p) = 0;

	/* change .* -> .s3b */
	c = (char *)mesh->textures;
	r = strlen(c);
	c[r-3] = 's';
	c[r-2] = '3';
	c[r-1] = 'b';

	fscanf(f, "%s", tmp);
      }

      /* numvert */
      if (strcmp(tmp, "numvert") != 0) {
	fprintf(stderr, "OBJECT_LOAD_AC3D: No numvert!\n");
	fclose(f);
	return FAILED;
      }
      fscanf(f, "%d", &mesh->vertices);

      mesh->vertice_list = malloc(sizeof(double) * 3 * mesh->vertices);
      if (mesh->vertice_list == NULL) {
	fprintf(stderr, "OBJECT_LOAD_AC3D: Out of memory error.\n");
	fclose(f);
	return FAILED;
      }
      d = mesh->vertice_list;

      /* load all vertices */
      for (i = 0; i < mesh->vertices; i++) {
	fscanf(f, "%f %f %f", &x, &y, &z);
	*d++ = x + loc_x;
	*d++ = y + loc_y;
	*d++ = z + loc_z;
      }

      /* numsurf */
      fscanf(f, "%s", tmp);
      if (strcmp(tmp, "numsurf") != 0) {
	fprintf(stderr, "OBJECT_LOAD_AC3D: No numsurf!\n");
	fclose(f);
	return FAILED;
      }
      fscanf(f, "%d", &mesh->triangles);
      skipped = 0;

      mesh->triangle_list = malloc(sizeof(int) * 3 * mesh->triangles);
      if (mesh->triangle_list == NULL) {
	fprintf(stderr, "OBJECT_LOAD_AC3D: Out of memory error.\n");
	fclose(f);
	return FAILED;
      }

      mesh->uv_list = malloc(sizeof(float) * 6 * mesh->triangles);
      if (mesh->uv_list == NULL) {
	fprintf(stderr, "OBJECT_LOAD_AC3D: Out of memory error.\n");
	fclose(f);
	return FAILED;
      }

      /* load surfaces (triangles) */
      e = mesh->uv_list;
      g = mesh->triangle_list;
      for (i = 0; i < mesh->triangles; i++) {
	/* SURF */
	fscanf(f, "%s", tmp);
	if (strcmp(tmp, "SURF") != 0) {
	  fprintf(stderr, "OBJECT_LOAD_AC3D: No SURF (only \"%s\")!\n", tmp);
	  fclose(f);
	  return FAILED;
	}

	/* skip surface flags */
	fscanf(f, "%*s");

	/* skip material */
	fscanf(f, "%*s %*s");

	/* refs */
	fscanf(f, "%s", tmp);
	if (strcmp(tmp, "refs") != 0) {
	  fprintf(stderr, "OBJECT_LOAD_AC3D: No refs!\n");
	  fclose(f);
	  return FAILED;
	}
	fscanf(f, "%d", &p);

	if (p != 3) {
	  int q;

	  fprintf(stderr, "OBJECT_LOAD_AC3D: Not a triangle! %d vertices! Skipping...\n", p);
	  skipped++;
	  for (q = 0; q < p; q++)
	    fscanf(f, "%*d %*f %*f");
	  continue;
	}

	fscanf(f, "%d %f %f", &p, &x, &y);
	*g++ = p;
	*e++ = x;
	*e++ = y;

	fscanf(f, "%d %f %f", &p, &x, &y);
	*g++ = p;
	*e++ = x;
	*e++ = y;

	fscanf(f, "%d %f %f", &p, &x, &y);
	*g++ = p;
	*e++ = x;
	*e++ = y;
      }

      /* add the mesh to the mesh pile */
      if (meshes_first == NULL) {
	meshes_first = mesh;
	meshes_last = mesh;
      }
      else {
	meshes_last->next = mesh;
	meshes_last = mesh;
      }
    }

    mesh->triangles -= skipped;

    /* kids */
    fscanf(f, "%s", tmp);
    if (strcmp(tmp, "kids") != 0) {
      fprintf(stderr, "OBJECT_LOAD_AC3D: No kids!\n");
      fclose(f);
      return FAILED;
    }

    fscanf(f, "%d", &p);
    if (p != 0) {
      fprintf(stderr, "OBJECT_LOAD_AC3D: More (%d) than zero kids!\n", p);
      fclose(f);
      return FAILED;
    }

    fscanf(f, "%s", tmp);
  }

  fclose(f);

  object_write_landscape(stdout, meshes_first);

  return SUCCEEDED;
}


int object_write_landscape(FILE *f, struct object_mesh *mesh) {

  double *a;
  float *c;
  int d, *b;


  fprintf(f, "// Landscape 3D ASCII\n");
  fprintf(f, "Meshes: %d\n", meshes_total);

  while (mesh != NULL) {
    fprintf(f, "// Mesh data\n");
    fprintf(f, "%s\n", mesh->name);
    fprintf(f, "Vertices: %d\n", mesh->vertices);
    fprintf(f, "Normals: NO\n");
    fprintf(f, "Triangles: %d\n", mesh->triangles);

    /* output vertice data */
    fprintf(f, "// Vertice data\n");
    a = mesh->vertice_list;
    for (d = 0; d < mesh->vertices; d++) {
      write_out_double(f, *(a+0), ' ');
      write_out_double(f, *(a+1), ' ');
      write_out_double(f, *(a+2), 0xA);
      a += 3;
    }

    /* output triangle data */
    fprintf(f, "// Triangle data\n");
    b = mesh->triangle_list;
    c = mesh->uv_list;
    for (d = 0; d < mesh->triangles; d++) {
      fprintf(f, "0 %d %d %d ", *(b+0), *(b+1), *(b+2));
      write_out_double(f, *(c+0), ' ');
      write_out_double(f, *(c+1), ' ');
      write_out_double(f, *(c+2), ' ');
      write_out_double(f, *(c+3), ' ');
      write_out_double(f, *(c+4), ' ');
      write_out_double(f, *(c+5), 0xA);
      b += 3;
      c += 6;
    }

    /* output texture data */
    fprintf(f, "// Texture data\n");
    fprintf(f, "Textures: 1\n");
    fprintf(f, "\"%s\"\n", (char *)mesh->textures);

    mesh = mesh->next;
  }

  return SUCCEEDED;
}


int write_out_double(FILE *f, double d, char c) {

  char tmp[32];
  int s;


  sprintf(tmp, "%f", d);
  s = strlen(tmp);

  if (s >= 32) {
    fprintf(stderr, "WRITE_OUT_DOUBLE: Too long a double (%d characters)... Imminent death.\n", s);
    exit(0);
  }

  s--;

  while (1) {
    if (tmp[s] == '0')
      tmp[s] = 0;
    else if (tmp[s] == '.') {
      s--;
      break;
    }
    else
      break;
    s--;
    if (s == 0)
      break;
  }
  s++;

  tmp[s++] = c;
  tmp[s] = 0;
  fprintf(f, tmp);

  return SUCCEEDED;
}
