
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "defines.h"
#include "common.h"
#include "build.h"
#include "../common/random.h"


extern struct plant **obj_plants;
extern struct psystem *psystems;
extern struct house *houses;
extern double *map, *map_vertices_z, *map_vertices_xy, random_xy;
extern float *map_normals, *map_triangle_normals, *map_fertility;
extern float *map_slopes;
extern char *map_shadow, *map_area, *map_triangle_area, *map_min_max;
extern int size, *map_triangles;

/* seed */
extern unsigned int seed;


/* remember to update MAP_GENERATOR_AREAS in defines.h */

int map_generator_build_areas(void) {

  int x, y;
  char *a, *b;


  /* set the block area type */
  for (x = 0; x < size * size; x++) {
    if (map_area[x] == MAP_AREA_UNDEFINED) {
      /* under the sea? */
      if (map[x] < LIMIT_SEA) {
	map_area[x] = MAP_AREA_UNDER_THE_SEA;
      }
      /* grasslands? */
      else if (map[x] < LIMIT_GRASS) {
	map_area[x] = MAP_AREA_GRASSLANDS;
      }
      /* mountains? */
      else {
	map_area[x] = MAP_AREA_MOUNTAIN;
      }
    }
  }

  /* set the triangle area type */
  a = map_area;
  b = map_triangle_area;
  for (y = 0; y < size; y++) {
    for (x = 0; x < size; x++) {
      *(b++) = *a;
      *(b++) = *a;
      *(b++) = *a;
      *(b++) = *a;
      a++;
    }
    a -= size;
    for (x = 0; x < size; x++) {
      *(b++) = *a;
      *(b++) = *a;
      *(b++) = *a;
      *(b++) = *a;
      a++;
    }
  }

  return SUCCEEDED;
}


int map_generator_compute_slopes(void) {

  double x1, x2, x3, y1, y2, y3, z1, z2, z3, dx1, dx2, dy1, dy2, dz1, dz2, ax, ay, az;
  int i, *r, n;


  r = map_triangles;
  for (i = 0; i < size*size*8; i++) {
    /* read the triangle */
    n = r[0];
    x1 = map_vertices_xy[n*2 + 0];
    y1 = map_vertices_xy[n*2 + 1];
    z1 = map_vertices_z[n];
    n = r[1];
    x2 = map_vertices_xy[n*2 + 0];
    y2 = map_vertices_xy[n*2 + 1];
    z2 = map_vertices_z[n];
    n = r[2];
    x3 = map_vertices_xy[n*2 + 0];
    y3 = map_vertices_xy[n*2 + 1];
    z3 = map_vertices_z[n];

    dx1 = x1 - x2;
    dy1 = y1 - y2;
    dz1 = z1 - z2;
    dx2 = x1 - x3;
    dy2 = y1 - y3;
    dz2 = z1 - z3;
    ax = dy1*dz2 - dy2*dz1;
    ay = dx2*dz1 - dx1*dz2;
    az = dx1*dy2 - dx2*dy1;

    dx1 = sqrt(ax*ax + ay*ay + az*az);
    ax /= dx1;
    ay /= dx1;
    az /= dx1;

    r += 3;
    map_slopes[i] = fabs(az);
  }

  return SUCCEEDED;
}


#define SLOPE_LIMIT 0.3


int map_generator_spread_minima(void) {

  double d;
  int x, y, i, n;


  i = 0;
  while (i == 0) {
    i = 1;
    for (y = 1; y < size-1; y++) {
      for (x = 1; x < size-1; x++) {
	if (map_min_max[y*size + x] == 0)
	  continue;
	n = y*size + x-1;
	if (map_min_max[n] == 0) {
	  d = fabs(map[y*size + x] - map[n]);
	  if (map_min_max[y*size + x] == MAP_MIN && d < SLOPE_LIMIT) {
	    i = 0;
	    map_min_max[n] = MAP_MIN;
	  }
	}
	n = (y-1)*size + x-1;
	if (map_min_max[n] == 0) {
	  d = fabs(map[y*size + x] - map[n]);
	  if (map_min_max[y*size + x] == MAP_MIN && d < SLOPE_LIMIT) {
	    i = 0;
	    map_min_max[n] = MAP_MIN;
	  }
	}
	n = (y-1)*size + x;
	if (map_min_max[n] == 0) {
	  d = fabs(map[y*size + x] - map[n]);
	  if (map_min_max[y*size + x] == MAP_MIN && d < SLOPE_LIMIT) {
	    i = 0;
	    map_min_max[n] = MAP_MIN;
	  }
	}
	n = (y-1)*size + x+1;
	if (map_min_max[n] == 0) {
	  d = fabs(map[y*size + x] - map[n]);
	  if (map_min_max[y*size + x] == MAP_MIN && d < SLOPE_LIMIT) {
	    i = 0;
	    map_min_max[n] = MAP_MIN;
	  }
	}
	n = y*size + x+1;
	if (map_min_max[n] == 0) {
	  d = fabs(map[y*size + x] - map[n]);
	  if (map_min_max[y*size + x] == MAP_MIN && d < SLOPE_LIMIT) {
	    i = 0;
	    map_min_max[n] = MAP_MIN;
	  }
	}
	n = (y+1)*size + x+1;
	if (map_min_max[n] == 0) {
	  d = fabs(map[y*size + x] - map[n]);
	  if (map_min_max[y*size + x] == MAP_MIN && d < SLOPE_LIMIT) {
	    i = 0;
	    map_min_max[n] = MAP_MIN;
	  }
	}
	n = (y+1)*size + x;
	if (map_min_max[n] == 0) {
	  d = fabs(map[y*size + x] - map[n]);
	  if (map_min_max[y*size + x] == MAP_MIN && d < SLOPE_LIMIT) {
	    i = 0;
	    map_min_max[n] = MAP_MIN;
	  }
	}
	n = (y+1)*size + x-1;
	if (map_min_max[n] == 0) {
	  d = fabs(map[y*size + x] - map[n]);
	  if (map_min_max[y*size + x] == MAP_MIN && d < SLOPE_LIMIT) {
	    i = 0;
	    map_min_max[n] = MAP_MIN;
	  }
	}
      }
    }
  }

  return SUCCEEDED;
}


int map_generator_find_local_minima(void) {

  double n[9];
  int x, y, min, max;


  min = 0;
  max = 0;
  for (y = 1; y < size-1; y++) {
    for (x = 1; x < size-1; x++) {
      n[0] = map[(y-1)*size + (x-1)];
      n[1] = map[(y-1)*size + (x-0)];
      n[2] = map[(y-1)*size + (x+1)];
      n[3] = map[(y-0)*size + (x-1)];
      n[4] = map[(y-0)*size + (x-0)];
      n[5] = map[(y-0)*size + (x+1)];
      n[6] = map[(y+1)*size + (x-1)];
      n[7] = map[(y+1)*size + (x-0)];
      n[8] = map[(y+1)*size + (x+1)];

      if (n[0] > n[4] && n[1] > n[4] && n[2] > n[4] &&
	  n[3] > n[4] && n[5] > n[4] &&
	  n[6] > n[4] && n[7] > n[4] && n[8] > n[4]) {
	map_min_max[y*size + x] = MAP_MIN;
	min++;
      }
    }
  }

  fprintf(stderr, "MAP_GENERATOR_FIND_LOCAL_MINIMA: %d local minima.\n", min);

  return SUCCEEDED;
}


int map_generator_compute_fertility(void) {

  float f;
  int x, y;


  for (y = 0; y < size; y++) {
    for (x = 0; x < size; x++) {
      if (map_min_max[y*size + x] == MAP_MIN)
	f = 2;
      else
	f = 1;
      f = f*(60 - map[y*size + x]);
      map_fertility[y*size + x] = f/40;
    }
  }

  return SUCCEEDED;
}


int map_generator_build_vertices(void) {

  double d, k, *q;
  int x, y, i, a, b, r;


  r = (size<<1) + 1;
  for (i = 0, y = 0; y < size; y++) {
    b = y<<1;
    for (x = 0; x < size; x++, i++) {
      a = x<<1;
      k = 0;
      map_read(x, y, &k);

      /* up row */
      d = k;
      map_read(x - 1, y - 1, &d);
      map_read(x - 1, y + 0, &d);
      map_read(x + 0, y - 1, &d);
      map_vertices_z[(b + 0)*r + a + 0] = d / 4;

      d = k;
      map_read(x + 0, y - 1, &d);
      map_vertices_z[(b + 0)*r + a + 1] = d / 2;

      d = k;
      map_read(x + 1, y - 1, &d);
      map_read(x + 1, y + 0, &d);
      map_read(x + 0, y - 1, &d);
      map_vertices_z[(b + 0)*r + a + 2] = d / 4;

      /* middle row */
      d = k;
      map_read(x - 1, y + 0, &d);
      map_vertices_z[(b + 1)*r + a + 0] = d / 2;

      map_vertices_z[(b + 1)*r + a + 1] = k;

      d = k;
      map_read(x + 1, y + 0, &d);
      map_vertices_z[(b + 1)*r + a + 2] = d / 2;

      /* bottom row */
      d = k;
      map_read(x - 1, y + 1, &d);
      map_read(x - 1, y + 0, &d);
      map_read(x + 0, y + 1, &d);
      map_vertices_z[(b + 2)*r + a + 0] = d / 4;

      d = k;
      map_read(x + 0, y + 1, &d);
      map_vertices_z[(b + 2)*r + a + 1] = d / 2;

      d = k;
      map_read(x + 1, y + 1, &d);
      map_read(x + 1, y + 0, &d);
      map_read(x + 0, y + 1, &d);
      map_vertices_z[(b + 2)*r + a + 2] = d / 4;
    }
  }

  q = map_vertices_xy;
  for (y = 0; y < r; y++) {
    for (x = 0; x < r; x++) {
      *(q++) = x; /* + random_gaussian(random_xy); */
      *(q++) = y; /* + random_gaussian(random_xy); */
    }
  }

  return SUCCEEDED;
}


int map_generator_build_triangles(void) {

  int *r, a, b, q;


  q = (size<<1) + 1;
  r = map_triangles;
  for (b = 0; b < (size<<1); b += 1) {
    for (a = 0; a < (size<<1); a += 2) {
      /* add eight new triangles (counter clockwise) per pixel */
      *(r++) = (b + 1)*q + 0 + a;
      *(r++) = (b + 1)*q + 1 + a; /* center */
      *(r++) = (b + 0)*q + 0 + a;

      *(r++) = (b + 0)*q + 0 + a;
      *(r++) = (b + 1)*q + 1 + a; /* center */
      *(r++) = (b + 0)*q + 1 + a;

      *(r++) = (b + 0)*q + 1 + a;
      *(r++) = (b + 1)*q + 1 + a; /* center */
      *(r++) = (b + 0)*q + 2 + a;

      *(r++) = (b + 0)*q + 2 + a;
      *(r++) = (b + 1)*q + 1 + a; /* center */
      *(r++) = (b + 1)*q + 2 + a;
    }
  }

  return SUCCEEDED;
}


static void _vector_normalize(float *x, float *y, float *z) {

  float d;


  d = sqrt((*x * *x) + (*y * *y) + (*z * *z));
  *x /= d;
  *y /= d;
  *z /= d;
}


int map_generator_build_normals(void) {

  float *f, fx, fy, fz;
  int x, y, *t, a, b, n;


  /* compute triangle face normals */
  t = map_triangles;
  f = map_triangle_normals;
  for (y = 0; y < (size<<1); y += 1) {
    for (x = 0; x < (size<<2); x += 4) {
      map_generator_compute_normal(*(t+0), *(t+1), *(t+2), f + ((y*size*4) + x + 0)*3);
      map_generator_compute_normal(*(t+3), *(t+4), *(t+5), f + ((y*size*4) + x + 1)*3);
      map_generator_compute_normal(*(t+6), *(t+7), *(t+8), f + ((y*size*4) + x + 2)*3);
      map_generator_compute_normal(*(t+9), *(t+10), *(t+11), f + ((y*size*4) + x + 3)*3);
      t += 12;
    }
  }

  /* compute vertex normals (average surrounding face normals) */
  f = map_normals;
  t = map_triangles;
  n = (size<<1) + 1;
  for (y = 0; y < n; y++) {
    for (x = 0; x < n; x++) {
      fx = 0;
      fy = 0;
      fz = 0;
      a = y;
      b = x<<1;
      if ((y & 1) == 0) {
	/* eight faces */
	if ((x & 1) == 0) {
	  map_generator_add_normal(b - 2, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b - 1, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b + 0, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b + 1, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b - 2, a + 0, &fx, &fy, &fz);
	  map_generator_add_normal(b - 1, a + 0, &fx, &fy, &fz);
	  map_generator_add_normal(b + 0, a + 0, &fx, &fy, &fz);
	  map_generator_add_normal(b + 1, a + 0, &fx, &fy, &fz);
	}
	/* four faces */
	else {
	  map_generator_add_normal(b - 1, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b + 0, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b - 1, a + 0, &fx, &fy, &fz);
	  map_generator_add_normal(b + 0, a + 0, &fx, &fy, &fz);
	}
      }
      else {
	/* eight faces */
	if ((x & 1) == 1) {
	  map_generator_add_normal(b - 2, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b - 1, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b + 0, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b + 1, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b - 2, a + 0, &fx, &fy, &fz);
	  map_generator_add_normal(b - 1, a + 0, &fx, &fy, &fz);
	  map_generator_add_normal(b + 0, a + 0, &fx, &fy, &fz);
	  map_generator_add_normal(b + 1, a + 0, &fx, &fy, &fz);
	}
	/* four faces */
	else {
	  map_generator_add_normal(b - 1, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b + 0, a - 1, &fx, &fy, &fz);
	  map_generator_add_normal(b - 1, a + 0, &fx, &fy, &fz);
	  map_generator_add_normal(b + 0, a + 0, &fx, &fy, &fz);
	}
      }
      _vector_normalize(&fx, &fy, &fz);
      *(f++) = fx;
      *(f++) = fy;
      *(f++) = fz;
    }
  }

  return SUCCEEDED;
}


int map_generator_add_normal(int x, int y, float *fx, float *fy, float *fz) {

  float *d;


  while (x < 0)
    x += size<<2;
  while (y < 0)
    y += size<<1;
  while (x >= (size<<2))
    x -= size<<2;
  while (y >= (size<<1))
    y -= size<<1;

  d = map_triangle_normals + (y*size*4*3) + x*3;
  *fx += *(d++);
  *fy += *(d++);
  *fz += *d;

  return SUCCEEDED;
}


int map_generator_compute_normal(int c1, int c2, int c3, float *d) {

  double dx1, dx2, dy1, dy2, dz1, dz2, r1, r2, r3;
  double x1, x2, x3, y1, y2, y3, z1, z2, z3;


  x1 = map_vertices_xy[(c1<<1) + 0];
  z1 = map_vertices_xy[(c1<<1) + 1];
  x2 = map_vertices_xy[(c2<<1) + 0];
  z2 = map_vertices_xy[(c2<<1) + 1];
  x3 = map_vertices_xy[(c3<<1) + 0];
  z3 = map_vertices_xy[(c3<<1) + 1];
  y1 = map_vertices_z[c1];
  y2 = map_vertices_z[c2];
  y3 = map_vertices_z[c3];

  dx1 = x1 - x2;
  dy1 = y1 - y2;
  dz1 = z1 - z2;
  dx2 = x1 - x3;
  dy2 = y1 - y3;
  dz2 = z1 - z3;
  r1 = dy1*dz2 - dy2*dz1;
  r2 = dx2*dz1 - dx1*dz2;
  r3 = dx1*dy2 - dx2*dy1;

  dx1 = sqrt(r1*r1 + r2*r2 + r3*r3);
  r1 /= dx1;
  r2 /= dx1;
  r3 /= dx1;

  /* swap y and z */
  *(d++) = r1;
  *(d++) = r3;
  *d = r2;

  return SUCCEEDED;
}


int map_generator_plant_psystems(void) {

  struct psystem *ps;
  double px, py, pz;
  int *d, i, x, y, p, z;


  d = map_triangles;
  for (i = 0, y = 0, p = 0; y < size*2; y++, p++) {
    for (x = 0; x < size; x++) {
      for (z = 0; z < 4; z++, i++) {
	if (map_triangle_area[i] == MAP_AREA_UNDER_THE_SEA) {
	  if ((rand()/(double)RAND_MAX) < 0.03) {
	    _get_point_inside_a_triangle(d, &px, &py, &pz);

	    ps = malloc(sizeof(struct psystem));
	    if (ps == NULL) {
	      fprintf(stderr, "MAP_GENERATOR_PLANT_PSYSTEMS: Out of memory error.\n");
	      return FAILED;
	    }

	    ps->x = px;
	    ps->y = pz;
	    ps->z = py;
	    ps->type = 0;
	    ps->rx = 0;
	    ps->ry = 1;
	    ps->rz = 0;

	    ps->next = psystems;
	    psystems = ps;
	  }
	}
	d += 3;
      }
    }
  }

  return SUCCEEDED;
}


int map_generator_plant_extras(void) {

  int i, *d, b, n, x, y, z, p, itr, iro, ibo, imu, m;
  double r;


  itr = 0;
  iro = 0;
  ibo = 0;
  imu = 0;

  d = map_triangles;
  for (i = 0, y = 0, p = 0; y < size*2; y++, p++) {
    for (x = 0; x < size; x++) {
      for (z = 0; z < 4; z++, i++) {
	if (map_triangle_area[i] == MAP_AREA_GRASSLANDS) {
	  /* plant a tree? */
	  m = (int)map_fertility[p*size + x];
	  /*
	  fprintf(stderr, "%d ", m);
	  */
	  if (map_slopes[i] < 0.9)
	    m /= 2;
	  if (map_slopes[i] < 0.8)
	    m /= 2;

	  for (b = 0; b < m; b++) {
	    if ((rand()/(double)RAND_MAX) < 0.03) {
	      itr++;
	      n = rand() % 3;
	      if (n == 0)
		plant_a_plant(i, d, OBJ_PINEWOOD_01);
	      else if (n == 1)
		plant_a_plant(i, d, OBJ_PINEWOOD_02);
	      else if (n == 2)
		plant_a_plant(i, d, OBJ_PINEWOOD_03);
	    }
	  }
	  /* plant mushrooms? */
	  r = rand()/(double)RAND_MAX;
	  if (r < 0.002) {
	    imu++;
	    plant_a_plant(i, d, OBJ_MUSHROOM_01);
	  }
	}
	/* plant rocks? */
	for (b = 0; b < 10; b++) {
	  r = rand()/(double)RAND_MAX;
	  if (r < 0.002) {
	    iro++;
	    plant_a_plant(i, d, OBJ_ROCK_01);
	  }
	}
	/* plant boulders? */
	r = rand()/(double)RAND_MAX;
	if (r < 0.002) {
	  ibo++;
	  plant_a_plant(i, d, OBJ_BOULDER_01);
	}
	d += 3;
      }
    }
    if ((y & 1) == 0)
      p--;
  }

  return SUCCEEDED;
}


int _get_point_inside_a_triangle(int *tri, double *x, double *y, double *z) {

  double f[3*3];
  int n;


  n = *(tri++);
  f[0*3 + 0] = map_vertices_xy[(n<<1)+0];
  f[0*3 + 1] = map_vertices_xy[(n<<1)+1];
  f[0*3 + 2] = map_vertices_z[n];
  n = *(tri++);
  f[1*3 + 0] = map_vertices_xy[(n<<1)+0];
  f[1*3 + 1] = map_vertices_xy[(n<<1)+1];
  f[1*3 + 2] = map_vertices_z[n];
  n = *tri;
  f[2*3 + 0] = map_vertices_xy[(n<<1)+0];
  f[2*3 + 1] = map_vertices_xy[(n<<1)+1];
  f[2*3 + 2] = map_vertices_z[n];

  random_point_inside_a_triangle(f, x, y, z);

  return SUCCEEDED;
}


int plant_a_plant(int i, int *d, int type) {

  struct plant *o;
  double x, y, z;


  _get_point_inside_a_triangle(d, &x, &y, &z);

  o = malloc(sizeof(struct plant));
  if (o == NULL) {
    fprintf(stderr, "PLANT_A_PLANT: Out of memory error.\n");
    exit(1);
  }

  o->type = type;
  o->x = x;
  o->y = y;
  o->z = z;
  o->rx = 0;
  o->ry = 360.0*rand()/(double)RAND_MAX;
  o->rz = 0;
  o->next = obj_plants[i];
  obj_plants[i] = o;

  return SUCCEEDED;
}


int map_generator_save_houses(void) {

  struct house *h;
  FILE *f;


  f = fopen("houses.txt", "wb");
  if (f == NULL) {
    fprintf(stderr, "MAP_GENERATOR_SAVE_HOUSES: Could not open file \"houses.txt\" for writing.\n");
    return FAILED;
  }

  fprintf(f, "%d\n", seed);

  h = houses;
  while (h != NULL) {
    fprintf(f, "%f ; %f ; %f ; %d ; %d\n", h->x, h->y, h->z, h->dx, h->dy);
    h = h->next;
  }

  fclose(f);

  return SUCCEEDED;
}


int map_generator_insert_houses(void) {

  struct house *h;
  int i;


  i = rand() % 20;
  while (i > 0) {
    h = malloc(sizeof(struct house));
    if (h == NULL) {
      fprintf(stderr, "MAP_GENERATOR_INSERT_HOUSES: Out of memory error.\n");
      return FAILED;
    }

    h->x = rand()/(double)RAND_MAX*MAP_PIXEL_SIDE_SCALE*size;
    h->y = rand()/(double)RAND_MAX*MAP_PIXEL_SIDE_SCALE*size;
    h->z = 1;
    h->dx = 4;
    h->dy = 10;
    h->next = houses;
    houses = h;

    i--;
  }

  return SUCCEEDED;
}
