
/* (c) 2003 ville helin <vhelin@cc.hut.fi> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "../common/defines.h"
#include "../common/lsload.h"
#include "../common/mesh.h"
#include "txt.h"
#include "simplify.h"


struct object_mesh *meshes_first = NULL, *meshes_last = NULL;


int mesh_normalize_textures(struct object_mesh *m, char *t, int s) {

  int i;


  if (m == NULL || t == NULL)
    return FAILED;

  /* remap the texture id's */
  for (i = 0; i < m->triangles; i++) {
    m->textures[i] = t[m->textures[i]] - 1;
  }

  return SUCCEEDED;
}


int mesh_collect_textures(struct object_mesh *m, char *t, int s) {

  int i, d;


  if (m == NULL || t == NULL)
    return FAILED;

  /* clear the buffer */
  for (i = 0; i < s; i++)
    t[i] = 0;

  /* mark the used textures */
  for (i = 0; i < m->triangles; i++) {
    d = m->textures[i];
    if (d >= s)
      fprintf(stderr, "MESH_COLLECT_TEXTURES: Texture ID overflow (%d).\n", d);
    else
      t[d] = 1;
  }

  /* renumber them */
  for (i = 0, d = 1; i < s; i++) {
    if (t[i] == 1) {
      t[i] = d;
      d++;
    }
  }

  return SUCCEEDED;
}


int main(int argc, char *argv[]) {

  struct txt_object *txt;
  struct object_mesh *m;
  float quality = 0;
  char t[256];
  int i, o;


  if (argc != 3) {
    printf("usage: %s <landscape file> <quality>\n", argv[0]);
    return 1;
  }

  /* parse quality */
  sscanf(argv[2], "%f", &quality);

  txt_init();
  object_load_show_errors(YES);

  /* load the object */
  if (object_load_landscape(argv[1]) == FAILED)
    return 1;

  m = meshes_first;
  i = 0;
  while (m != NULL) {
    i++;
    m = m->next;
  }

  printf("// Landscape 3D ASCII\n");
  printf("Meshes: %d\n", i);

  /* simplify the meshes */
  m = meshes_first;
  while (m != NULL) {

    /* generate missing normals */
    if (m->face_normal_list == NULL)
      mesh_compute_face_normals(m);
    if (m->vertex_normal_list == NULL)
      mesh_compute_vertex_normals(m);

    simplify_mesh(m, quality);

    /* recompute the normals */
    mesh_compute_face_normals(m);
    mesh_compute_vertex_normals(m);

    mesh_collect_textures(m, t, 256);
    mesh_normalize_textures(m, t, 256);
    mesh_write(m, stdout);

    /* count the textures */
    for (i = 0, o = 0; i < 256; i++) {
      if (t[i] > 0)
	o++;
    }

    printf("// Texture data\n");
    printf("Textures: %d\n", o);

    /* print the textures */
    for (i = 0, o = 0; i < 256; i++) {
      if (t[i] > 0) {
	txt_get(i, &txt);
	printf("\"%s\"\n", txt->name);
      }
    }

    m = m->next;
  }

  return 0;
}
