
/* (c) 2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "../common/defines.h"
#include "../common/lsload.h"
#include "../common/mesh.h"
#include "../common/mesh2.h"
#include "main.h"
#include "txt.h"


extern int textures;

struct object_mesh *meshes_first = NULL, *meshes_last = NULL;
struct mesh2 *m2 = NULL;


int main(int argc, char *argv[]) {

  struct object_mesh *m;


  if (argc != 3) {
    fprintf(stderr, "usage: %s <landscape file> <mesh2 file>\n", argv[0]);
    return 1;
  }

  txt_init();

  if (object_load_landscape(argv[1]) == FAILED)
    return 1;

  m = meshes_first;
  while (m != NULL) {
    /* generate missing normals */
    if (m->face_normal_list == NULL)
      mesh_compute_face_normals(m);
    if (m->vertex_normal_list == NULL)
      mesh_compute_vertex_normals(m);
    m = m->next;
  }

  /* do the conversion */
  if (landscape_to_mesh2(meshes_first, &m2) == FAILED)
    return 1;

  /* write out mesh2 */
  mesh2_write_binary(argv[2], m2);

  return 0;
}


int landscape_to_mesh2(struct object_mesh *l, struct mesh2 **m) {

  char memerr[] = "LANDSCAPE_TO_MESH2: Out of memory error.\n";
  struct object_mesh *o;
  struct mesh2 *t;
  int i, n, x;


  t = malloc(sizeof(struct mesh2));
  if (t == NULL) {
    fprintf(stderr, memerr);
    return FAILED;
  }

  /* merge vertex lists */
  i = 0;
  o = l;
  x = 0;
  while (o != NULL) {
    i += o->triangles;
    o = o->next;
    x++;
  }

  fprintf(stderr, "MESHES: %d\n", x);
  fprintf(stderr, "TRIANGLES: %d\n", i);
  t->triangles = i;

  t->vertex_list = malloc(sizeof(float)*3*3*i);
  t->vertex_normal_list = malloc(sizeof(float)*3*3*i);
  t->face_normal_list = malloc(sizeof(float)*3*i);
  t->uv_list = malloc(sizeof(float)*2*3*i);
  if (t->vertex_list == NULL || t->vertex_normal_list == NULL ||
      t->face_normal_list == NULL || t->uv_list == NULL) {
    fprintf(stderr, memerr);
    return FAILED;
  }

  t->vertices = 3*i;

  /* generate triangle lists for each texture */
  fprintf(stderr, "TEXTURES: %d\n", textures);
  t->objects = textures;

  /* count the triangles */
  for (i = 0; i < textures; i++) {
    n = 0;
    o = l;
    while (o != NULL) {
      for (x = 0; x < o->triangles; x++) {
	if (o->textures[x] == i)
	  n++;
      }
      o = o->next;
    }

    fprintf(stderr, "TEXTURE %d - %d TRIANGLES\n", i, n);

    t->object_data[i*2 + 0] = i;
    t->object_data[i*2 + 1] = n;
  }

  /* expand the data */
  n = 0;
  for (i = 0; i < textures; i++) {
    o = l;
    while (o != NULL) {
      for (x = 0; x < o->triangles; x++) {
	if (o->textures[x] == i) {
	  /* add the triangle to the list */
	  t->vertex_list[n*3*3 + 0] = o->vertice_list[o->triangle_list[x*3 + 0]*3 + 0];
	  t->vertex_list[n*3*3 + 1] = o->vertice_list[o->triangle_list[x*3 + 0]*3 + 1];
	  t->vertex_list[n*3*3 + 2] = o->vertice_list[o->triangle_list[x*3 + 0]*3 + 2];
	  t->vertex_list[n*3*3 + 3] = o->vertice_list[o->triangle_list[x*3 + 1]*3 + 0];
	  t->vertex_list[n*3*3 + 4] = o->vertice_list[o->triangle_list[x*3 + 1]*3 + 1];
	  t->vertex_list[n*3*3 + 5] = o->vertice_list[o->triangle_list[x*3 + 1]*3 + 2];
	  t->vertex_list[n*3*3 + 6] = o->vertice_list[o->triangle_list[x*3 + 2]*3 + 0];
	  t->vertex_list[n*3*3 + 7] = o->vertice_list[o->triangle_list[x*3 + 2]*3 + 1];
	  t->vertex_list[n*3*3 + 8] = o->vertice_list[o->triangle_list[x*3 + 2]*3 + 2];
	  t->vertex_normal_list[n*3*3 + 0] = o->vertex_normal_list[o->triangle_list[x*3 + 0]*3 + 0];
	  t->vertex_normal_list[n*3*3 + 1] = o->vertex_normal_list[o->triangle_list[x*3 + 0]*3 + 1];
	  t->vertex_normal_list[n*3*3 + 2] = o->vertex_normal_list[o->triangle_list[x*3 + 0]*3 + 2];
	  t->vertex_normal_list[n*3*3 + 3] = o->vertex_normal_list[o->triangle_list[x*3 + 1]*3 + 0];
	  t->vertex_normal_list[n*3*3 + 4] = o->vertex_normal_list[o->triangle_list[x*3 + 1]*3 + 1];
	  t->vertex_normal_list[n*3*3 + 5] = o->vertex_normal_list[o->triangle_list[x*3 + 1]*3 + 2];
	  t->vertex_normal_list[n*3*3 + 6] = o->vertex_normal_list[o->triangle_list[x*3 + 2]*3 + 0];
	  t->vertex_normal_list[n*3*3 + 7] = o->vertex_normal_list[o->triangle_list[x*3 + 2]*3 + 1];
	  t->vertex_normal_list[n*3*3 + 8] = o->vertex_normal_list[o->triangle_list[x*3 + 2]*3 + 2];
	  t->face_normal_list[n*3 + 0] = o->face_normal_list[x*3 + 0];
	  t->face_normal_list[n*3 + 1] = o->face_normal_list[x*3 + 1];
	  t->face_normal_list[n*3 + 2] = o->face_normal_list[x*3 + 2];
	  t->uv_list[n*3*2 + 0] = o->uv_list[x*3*2 + 0];
	  t->uv_list[n*3*2 + 1] = o->uv_list[x*3*2 + 1];
	  t->uv_list[n*3*2 + 2] = o->uv_list[x*3*2 + 2];
	  t->uv_list[n*3*2 + 3] = o->uv_list[x*3*2 + 3];
	  t->uv_list[n*3*2 + 4] = o->uv_list[x*3*2 + 4];
	  t->uv_list[n*3*2 + 5] = o->uv_list[x*3*2 + 5];
	  n++;
	}
      }
      o = o->next;
    }
  }

  *m = t;

  return SUCCEEDED;
}
