
/* (c) 2001 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "defines.h"
#include "main.h"

/*
#define DEBUG 1
*/

struct joint *joint_first = NULL;
int joints = 0;


int main(int argc, char *argv[]) {

  struct joint *j;
  double *m;
  char *d;
  FILE *f;
  int s, n;


  if (argc != 3) {
    fprintf(stderr, "usage: %s <joint module> <anim name>\n", argv[0]);
    return 1;
  }

  f = fopen(argv[1], "rb");
  if (f == NULL) {
    fprintf(stderr, "MAIN: Could not open file \"%s\".\n", argv[1]);
    return 1;
  }

  fseek(f, 0, SEEK_END);
  s = ftell(f);
  fseek(f, 0, SEEK_SET);

  d = malloc(s);
  if (d == NULL) {
    fprintf(stderr, "MAIN: Out of memory.\n");
    fclose(f);
    return FAILED;
  }

  fread(d, 1, s, f);
  fclose(f);

  if (parse_file(d, s) == FAILED) {
    free(d);
    return 1;
  }

  compute_joint_matrices();
  compute_global_joint_matrices();

  free(d);

  if (joint_first == NULL) {
    fprintf(stderr, "MAIN: No joints?\n");
    return 1;
  }

  /* write global joint matrices */
  f = fopen(argv[2], "wb");
  if (f == NULL) {
    fprintf(stderr, "MAIN: Could not open file \"%s\".\n", argv[2]);
    return 1;
  }

  fprintf(f, "Frames: %d\nJoints: %d\n", joint_first->frames, joints);
  j = joint_first;
  n = 0;
  while (j != NULL) {
    fprintf(f, "\nJoint: %d\n", n);
    m = j->global_joint_matrices;
    for (s = 0; s < j->frames; s++) {
      /* write the 4x4 global joint matrix row by row */
      fprintf(f, "\n%f\t%f\t%f\t%f\n", *m, *(m+1), *(m+2), *(m+3));
      m += 4;
      fprintf(f, "%f\t%f\t%f\t%f\n", *m, *(m+1), *(m+2), *(m+3));
      m += 4;
      fprintf(f, "%f\t%f\t%f\t%f\n", *m, *(m+1), *(m+2), *(m+3));
      m += 4;
      /* _if_ the computations are correct, the bottom row will be { 0, 0, 0, 1 } */
      /* so it's waste of space to write them out                                 */
#ifdef DEBUG
      fprintf(f, "%f\t%f\t%f\t%f\n", *m, *(m+1), *(m+2), *(m+3));
#endif
      if (*m != 0.0 || *(m+1) != 0.0 || *(m+2) != 0.0 || *(m+3) != 1.0)
	fprintf(stderr, "MAIN: Error in matrix.\n");
      m += 4;
    }
    n++;
    j = j->next;
  }

  fclose(f);

  return 0;
}


int compute_joint_matrices(void) {

  struct joint *j;
  double m[4 * 4];
  int i;


  j = joint_first;
  while (j != NULL) {
    create_joint_matrix(j->ax, j->ay, j->az, j->x, j->y, j->z, j->joint_matrix);
    for (i = 0; i < j->frames; i++) {
      create_joint_matrix(j->rotation[i*3], j->rotation[i*3+1], j->rotation[i*3+2], j->position[i*3], j->position[i*3+1], j->position[i*3+2], m);
      matrix_4x4_multiply(j->joint_matrix, m, &(j->global_joint_matrices[i*4*4]));
    }
    j = j->next;
  }

  return SUCCEEDED;
}


int compute_global_joint_matrices(void) {

  struct joint *j, *t, *m;
  double tm[4*4];
  int i;


  j = joint_first;
  while (j != NULL) {
    if (j->parent[0] == 0) {
      j = j->next;
      continue;
    }

    m = j;
    t = joint_first;
    while (m->parent[0] != 0) {
      if (strcmp(t->name, m->parent) == 0) {
	for (i = 0; i < m->frames; i++) {
	  matrix_4x4_multiply(&(j->global_joint_matrices[i*4*4]), &(t->global_joint_matrices[i*4*4]), tm);
	  memcpy(&(j->global_joint_matrices[i*4*4]), tm, sizeof(double)*4*4);
	}
	m = t;
	t = joint_first;
      }
      t = t->next;
    }

    j = j->next;
  }

  return SUCCEEDED;
}


int create_joint_matrix(double ax, double ay, double az, double x, double y, double z, double *jm) {

  double sr, sp, sy, cr, cp, cy;


  sy = sin(az);
  cy = cos(az);
  sp = sin(ay);
  cp = cos(ay);
  sr = sin(ax);
  cr = cos(ax);
  jm[0 * 4 + 0] = cp * cy;
  jm[1 * 4 + 0] = cp * sy;
  jm[2 * 4 + 0] = -sp;
  jm[0 * 4 + 1] = sr * sp * cy+cr * (-sy);
  jm[1 * 4 + 1] = sr * sp * sy+cr * cy;
  jm[2 * 4 + 1] = sr * cp;
  jm[0 * 4 + 2] = cr * sp * cy + (-sr) * (-sy);
  jm[1 * 4 + 2] = cr * sp * sy + (-sr) * cy;
  jm[2 * 4 + 2] = cr * cp;
  jm[0 * 4 + 3] = x;
  jm[1 * 4 + 3] = y;
  jm[2 * 4 + 3] = z;
  jm[3 * 4 + 0] = 0.0;
  jm[3 * 4 + 1] = 0.0;
  jm[3 * 4 + 2] = 0.0;
  jm[3 * 4 + 3] = 1.0;

  return SUCCEEDED;
}


int parse_file(char *b, int s) {

  struct joint *a;
  double d, keys_double_list[MAX_SIZE*3];
  char r[256];
  int i = 0, y, q, frames, keys, n, keys_time_list[MAX_SIZE];


  while (i < s) {
    object_get_next_token(b, r, &i, &y, s);
    /* frames? */
    if (strcmp(r, "Frames:") == 0)
      break;
  }

  if (i >= s) {
    fprintf(stderr, "PARSE_FILE: Invalid input file.\n");
    return FAILED;
  }

  object_get_next_token(b, r, &i, &y, s);
  obtain_integer(r, &frames);

  while (i < s) {
    object_get_next_token(b, r, &i, &y, s);
    /* bones? */
    if (strcmp(r, "Bones:") == 0)
      break;
  }

  if (i >= s) {
    fprintf(stderr, "PARSE_FILE: Invalid input file.\n");
    return FAILED;
  }

  object_get_next_token(b, r, &i, &y, s);
  obtain_integer(r, &joints);

  for (q = 0; q < joints; q++) {
    a = malloc(sizeof(struct joint));
    if (a == NULL) {
      fprintf(stderr, "PARSE_FILE: Out of memory.\n");
      return FAILED;
    }

    /* joint name */
    object_get_next_token(b, a->name, &i, &y, s);
    /* parent joint name */
    object_get_next_token(b, a->parent, &i, &y, s);
    /* skip flags */
    object_get_next_token(b, r, &i, &y, s);
    /* x */
    object_get_next_token(b, r, &i, &y, s);
    obtain_double(r, &a->x);
    /* y */
    object_get_next_token(b, r, &i, &y, s);
    obtain_double(r, &a->y);
    /* z */
    object_get_next_token(b, r, &i, &y, s);
    obtain_double(r, &a->z);
    /* ax */
    object_get_next_token(b, r, &i, &y, s);
    obtain_double(r, &a->ax);
    /* ay */
    object_get_next_token(b, r, &i, &y, s);
    obtain_double(r, &a->ay);
    /* az */
    object_get_next_token(b, r, &i, &y, s);
    obtain_double(r, &a->az);

    /* amount of position keys */
    object_get_next_token(b, r, &i, &y, s);
    obtain_integer(r, &keys);

    for (n = 0; n < keys; n++) {
      /* time */
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &d);
      keys_time_list[n] = (int)d - 1;
      /* x */
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &keys_double_list[n*3+0]);
      /* y */
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &keys_double_list[n*3+1]);
      /* z */
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &keys_double_list[n*3+2]);
    }

    interpolate_data(keys, frames, keys_time_list, keys_double_list, a->position);

    /* amount of rotation keys */
    object_get_next_token(b, r, &i, &y, s);
    obtain_integer(r, &keys);

    for (n = 0; n < keys; n++) {
      /* time */
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &d);
      keys_time_list[n] = (int)d - 1;
      /* ax */
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &keys_double_list[n*3+0]);
      /* ay */
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &keys_double_list[n*3+1]);
      /* az */
      object_get_next_token(b, r, &i, &y, s);
      obtain_double(r, &keys_double_list[n*3+2]);
    }

    interpolate_data(keys, frames, keys_time_list, keys_double_list, a->rotation);

    a->frames = frames;
    if (joint_first == NULL) {
      a->next = NULL;
      joint_first = a;
    }
    else {
      a->next = joint_first;
      joint_first = a;
    }
  }

  return SUCCEEDED;
}


int interpolate_data(int keys, int frames, int *keys_time_list, double *keys_double_list, double *out) {

  int i, a, b, q, p;
  double xa, ya, za, xt, yt, zt, xe, ye, ze;


  if (keys < 1) {
    fprintf(stderr, "INTERPOLATE_DATA: Not enough keys (%d).\n", keys);
    return FAILED;
  }

  /* just copy the frames */
  if (keys == 1) {
    for (i = 0; i < frames; i++)
      memcpy(&(out[i * 3]), &(keys_double_list[i * 3]), sizeof(double) * 3);
    return SUCCEEDED;
  }

  if (keys_time_list[0] != 0 || keys_time_list[keys - 1] != frames - 1) {
    fprintf(stderr, "INTERPOLATE_DATA: Animation %d->%d, not what FRAMES specified 0->%d.\n", keys_time_list[0], keys_time_list[keys - 1], frames - 1);
    return FAILED;
  }

  a = keys_time_list[0];
  xa = keys_double_list[0];
  ya = keys_double_list[1];
  za = keys_double_list[2];
  for (i = 1; i < keys; i++) {
    b = keys_time_list[i];
    xe = keys_double_list[i * 3];
    ye = keys_double_list[i * 3 + 1];
    ze = keys_double_list[i * 3 + 2];
    xt = (xe - xa) / (b - a);
    yt = (ye - ya) / (b - a);
    zt = (ze - za) / (b - a);
    for (q = a, p = 0; q <= b; q++, p++) {
      out[q * 3 + 0] = xa + (xt * p);
      out[q * 3 + 1] = ya + (yt * p);
      out[q * 3 + 2] = za + (zt * p);
    }
    xa = xe;
    ya = ye;
    za = ze;
    a = b;
  }

  return SUCCEEDED;
}


int obtain_double(char *b, double *d) {

  int i, m, c;


  if (*b == '-') {
    m = -1;
    b++;
  }
  else
    m = 1;

  i = 0;
  while (*b != 0 && *b != '.') {
    i = i * 10 + *b - '0';
    b++;
  }

  if (*b == 0) {
    *d = i * m;
    return SUCCEEDED;
  }

  b++;
  *d = 0;
  c = 10;
  while (*b != 0) {
    *d += ((double)(*b - '0')) / c;
    b++;
    c *= 10;
  }

  *d = (i + *d) * m;
  return SUCCEEDED;
}


int obtain_integer(char *b, int *i) {

  int m;


  if (*b == '-') {
    m = -1;
    b++;
  }
  else
    m = 1;

  *i = 0;
  while (*b != 0) {
    *i = *i * 10 + *b - '0';
    b++;
  }

  *i = *i * m;

  return SUCCEEDED;
}


int object_get_next_token(char *b, char *r, int *s, int *k, int m) {

  while (*s < m && (*(b+*s) == ' ' || *(b+*s) == 0x0a || *(b+*s) == 0x0d))
    (*s)++;

  *k = 0;

  if (*(b+*s) == '"') {
    (*s)++;
    while (*s < m && *(b+*s) != '"') {
      *(r+*k) = *(b+*s);
      (*s)++;
      (*k)++;
    }
    (*s)++;
  }
  else {
    while (*s < m && *(b+*s) != ' ' && *(b+*s) != 0x0a && *(b+*s) != 0x0d) {
      *(r+*k) = *(b+*s);
      (*s)++;
      (*k)++;
    }
  }

  *(r+*k) = 0;

  return SUCCEEDED;
}


int matrix_4x4_multiply(double *a, double *b, double *out) {

  int x, y;


  /* compute out = a * b */
  for (y = 0; y < 4; y++)
    for (x = 0; x < 4; x++)
      out[y*4 + x] = a[y*4] * b[x] + a[y*4 + 1] * b[4 + x] + a[y*4 + 2] * b[8 + x] + a[y*4 + 3] * b[12 + x];

  return SUCCEEDED;
}
