
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "game_object.h"
#include "main.h"
#include "id.h"
#include "../common/parse.h"


extern unsigned int current_time, *world_time;
extern int world_x, world_y, world_z;
extern char world_name[256];

struct game_object **game_object_map = NULL, **game_object_table = NULL;


int game_object_init(void) {

  int i;


  game_object_map = malloc(sizeof(struct game_object *) * world_x*world_y*world_z);
  game_object_table = malloc(sizeof(struct game_object *) * MAX_ID);
  if (game_object_map == NULL || game_object_table == NULL) {
    fprintf(stderr, "GAME_OBJECT_INIT: Out of memory error.\n");
    return FAILED;
  }

  for (i = 0; i < world_x*world_y*world_z; i++)
    game_object_map[i] = NULL;

  for (i = 0; i < MAX_ID; i++)
    game_object_table[i] = NULL;

  return SUCCEEDED;
}


struct game_object *game_object_get(unsigned int id) {

  if (id < 0 || id >= MAX_ID) {
    fprintf(stderr, "GAME_OBJECT_GET: ID %d out of range.\n", id);
    return NULL;
  }

  if (game_object_table[id] == NULL)
    fprintf(stderr, "GAME_OBJECT_GET: Missing object %d.\n", id);

  return game_object_table[id];
}


void game_object_remove(struct game_object *g) {

  char tmp[8];
  int i;


  if (g->id < 0 || g->id > MAX_ID) {
    fprintf(stderr, "GAME_OBJECT_REMOVE: ID %d is out of range.\n", g->id);
    return;
  }

  if (g->prev == NULL) {
    i = g->cz*world_y*world_x + g->cy*world_x + g->cx;
    if (game_object_map[i] != g) {
      fprintf(stderr, "GAME_OBJECT_REMOVE: Mangled object %d.\n", g->id);
      return;
    }
    game_object_map[i] = g->next;
  }

  game_object_table[g->id] = NULL;

  if (g->next != NULL)
    g->next->prev = g->prev;
  if (g->prev != NULL)
    g->prev->next = g->next;

  /* remove message */
  tmp[0] = 'X';
  write_int(tmp + 1, g->id);

  send_to_all(tmp, 5);
  free_id(g->id);
  free(g);
}


void game_object_add(struct game_object *g) {

  int x, y, z, i;


  if (g->id < 0 || g->id > MAX_ID) {
    fprintf(stderr, "GAME_OBJECT_ADD: ID %d is out of range.\n", g->id);
    return;
  }

  if (game_object_table[g->id] != NULL)
    fprintf(stderr, "GAME_OBJECT_ADD: Duplicate ID %d.\n", g->id);
  game_object_table[g->id] = g;

  x = g->x / CUBE_SIDE;
  y = g->y / CUBE_SIDE;
  z = g->z / CUBE_SIDE;

  if (x < 0 || x >= world_x)
    x = 0;
  if (y < 0 || y >= world_y)
    y = 0;
  if (z < 0 || z >= world_z)
    z = 0;

  g->cx = x;
  g->cy = y;
  g->cz = z;

  i = z*world_y*world_x + y*world_x + x;

  g->next = game_object_map[i];
  g->prev = NULL;
  if (game_object_map[i] != NULL)
    game_object_map[i]->prev = g;
  game_object_map[i] = g;

  /* mark the cube as "dirty" */
  world_time[i] = current_time;
  g->change_time = current_time;
}


int game_object_load_instances(void) {

  unsigned int id;
  double x, y, z;
  float ax, ay, az;
  FILE *f;
  int i, d;
  char a[256];


  strcpy(a, "../data/wld/");
  strcat(a, world_name);
  strcat(a, "/plants.txt");

  f = fopen(a, "rb");
  if (f == NULL) {
    fprintf(stderr, "GAME_OBJECT_LOAD_INSTANCES: Could not open \"%s\".\n", a);
    return FAILED;
  }

  d = 0;
  while (fscanf(f, "%lf %lf %lf %f %f %f %d", &x, &y, &z, &ax, &ay, &az, &i) != EOF) {
    if (generate_id(&id) == FAILED) {
      fclose(f);
      return FAILED;
    }
    sprintf(a, "%d%c", id, 0);
    game_object_create(a, id, OBJECT_OTHER, i, x, y, z, ax, ay, az, NULL, YES);
    d++;
  }

  fprintf(stderr, "GAME_OBJECT_LOAD_INSTANCES: Loaded %d objects.\n", d);
  fclose(f);

  return SUCCEEDED;
}


int game_object_create(char *name, unsigned int id, int type, int object, double x, double y, double z, float ax, float ay, float az, struct game_object **o, int c) {

  struct game_object *g;


  g = malloc(sizeof(struct game_object));
  if (g == NULL) {
    fprintf(stderr, "GAME_OBJECT_CREATE: Out of memory error.\n");
    return FAILED;
  }

  if (name != NULL)
    strcpy(g->name, name);
  else
    g->name[0] = 0;

  g->id = id;
  g->type = type;
  g->object = object;
  g->x = x;
  g->y = y;
  g->z = z;
  g->ax = ax;
  g->ay = ay;
  g->az = az;

  /* is the object collidable? */
  if (c == YES)
    g->status = GO_COLLISION;
  else
    g->status = 0;

  game_object_add(g);

  if (o != NULL)
    *o = g;

  return SUCCEEDED;
}


void game_object_relocate(struct game_object *g) {

  unsigned int x, y, z, i;


  x = g->x / CUBE_SIDE;
  y = g->y / CUBE_SIDE;
  z = g->z / CUBE_SIDE;

  if (x < 0 || x >= world_x)
    x = 0;
  if (y < 0 || y >= world_y)
    y = 0;
  if (z < 0 || z >= world_z)
    z = 0;

  /* is the object still in the right cube? */
  if (g->cx == x && g->cy == y && g->cz == z)
    return;

  /* cube changed -> relocate the object */
  if (g->next != NULL)
    g->next->prev = g->prev;
  if (g->prev != NULL)
    g->prev->next = g->next;
  if (g->prev == NULL) {
    i = g->cz*world_y*world_x + g->cy*world_x + g->cx;
    game_object_map[i] = g->next;
  }

  g->cx = x;
  g->cy = y;
  g->cz = z;    

  i = z*world_y*world_x + y*world_x + x;
  g->prev = NULL;
  g->next = game_object_map[i];
  if (game_object_map[i] != NULL)
    game_object_map[i]->prev = g;
  game_object_map[i] = g;
}


