
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>
#include <sched.h>

#include "defines.h"
#include "player.h"
#include "main.h"
#include "id.h"
#include "game_object.h"
#include "psystems.h"
#include "../common/socket.h"
#include "../common/parse.h"
#include "../common/dynmem.h"
#include "../common/configuration.h"


struct player *players = NULL;
extern struct game_object **world_objects, **game_object_table;
fd_set active_fd_set, read_fd_set, passwd_fd_set, new_fd_set, passwd_read_fd_set, data_fd_set, player_fd_set;
char session_passwd[256], master_passwd[256], world_name[256];
int main_port = 6789, main_socket, main_port_active = OFF;
int master_port, master_socket, accept_clients = YES;
int priority = 0;

int world_x, world_y, world_z;
unsigned int current_time, *world_time, world_map, world_seed;

volatile int master_active = OFF, server_status = ON;

/* buffer for reading data from the players */
extern char *comm_buff;
extern int comm_buff_size;

/* server's configuration file */
struct configuration configuration_server[] = {
  { "session_passwd:", CONFIGURATION_TYPE_STRING, session_passwd },
  { "master_passwd:", CONFIGURATION_TYPE_STRING, master_passwd },
  { "world_name:", CONFIGURATION_TYPE_STRING, world_name },
  { "priority:", CONFIGURATION_TYPE_INTEGER, &priority },
  { NULL, 0, NULL }
};



int main(int argc, char *argv[]) {

  srand(time(NULL));

  /* set the server's priority */
  setpriority(PRIO_PROCESS, 0, 20);

  if (signal(SIGINT, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGKILL, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGTERM, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGSEGV, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGFPE, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGILL, signal_handler_int) == SIG_IGN)
    return 1;

  if (initialize_id_pool() == FAILED)
    return 1;

  /* set the defaults */
  strcpy(session_passwd, "candy!");
  strcpy(master_passwd, "sugar!");
  strcpy(world_name, "2");

  /* read server.cfg */
  configuration_read("server.cfg", configuration_server);

  /* read world.txt */
  if (read_world_cfg() == FAILED)
    return 1;

  /* set the priority */
  setpriority(PRIO_PROCESS, 0, priority);

  /* set the initial input buffer size */
  dynmem_handle_buffer(&comm_buff, &comm_buff_size, 64*1024);

  current_time = 1;

  world_init();

  /* game objects */
  game_object_init();
  game_object_load_instances();

  /* particle systems */
  psystems_init();
  psystems_load_instances();

  if (socket_create(&main_socket, main_port, YES) == FAILED)
    return 1;
  main_port_active = ON;

  /* main loop */
  handle_sockets();

  close_all_sockets();

  return 0;
}


int read_world_cfg(void) {

  char tmp[256], c[256];
  FILE *f;


  strcpy(c, "../data/wld/");
  strcat(c, world_name);
  strcat(c, "/world.txt");

  f = fopen(c, "rb");
  if (f == NULL) {
    fprintf(stderr, "READ_WORLD_CFG: Could not open file \"%s\" for reading.\n", c);
    return FAILED;
  }

  while (fscanf(f, "%255s", tmp) != EOF) {
    if (strcmp(tmp, "MAP:") == 0) {
      fscanf(f, "%255s", tmp);
      if (strcmp(tmp, "GENERATED") == 0)
	world_map = WORLD_MAP_GENERATED;
      else if (strcmp(tmp, "IMAGE") == 0)
	world_map = WORLD_MAP_IMAGE;
      else
	fprintf(stderr, "READ_WORLD_CFG: File \"%s\", unknown symbol \"%s\".\n", c, tmp);
    }
    else if (strcmp(tmp, "SEED:") == 0) {
      fscanf(f, "%u", &world_seed);
    }
    else {
      fprintf(stderr, "READ_WORLD_CFG: File \"%s\", unknown symbol \"%s\".\n", c, tmp);
    }
  }

  fclose(f);

  return SUCCEEDED;
}


int handle_sockets(void) {

  struct sockaddr_in clientname;
  struct player *p;
  struct timeval t;
  size_t size;
  char b[1024];
  int i, s, po, so, n, x;


  /* initialize the set of active sockets */
  FD_ZERO(&active_fd_set);
  FD_ZERO(&passwd_fd_set);
  FD_ZERO(&passwd_read_fd_set);
  FD_ZERO(&data_fd_set);
  FD_ZERO(&new_fd_set);
  FD_ZERO(&player_fd_set);
  FD_SET(main_socket, &new_fd_set);
  FD_SET(main_socket, &active_fd_set);
  FD_SET(main_socket, &passwd_fd_set);

  while (server_status == ON) {
    read_fd_set = active_fd_set;
    t.tv_sec = 0;
    t.tv_usec = 1;
    i = select(FD_SETSIZE, &read_fd_set, NULL, NULL, &t);
    if (i < 0) {
      fprintf(stderr, "HANDLE_SOCKETS: Select failed.\n");
      socket_delete(main_socket);
      _exit(EXIT_FAILURE);
    }
    if (i == 0) {
      /* no new connections -> manage the old ones */
      player_carousel();
      continue;
    }

    for (i = 0; i < FD_SETSIZE; i++) {
      if (FD_ISSET(i, &read_fd_set)) {
	/* connection request */
	if (FD_ISSET(i, &new_fd_set)) {
	  size = sizeof(struct sockaddr_in);
	  s = accept(i, (struct sockaddr *)&clientname, &size);
	  if (s < 0) {
	    fprintf(stderr, "HANDLE_SOCKETS: Accept failed.\n");
	    socket_delete(main_socket);
	    _exit(EXIT_FAILURE);
	  }
	  fprintf(stderr, "HANDLE_SOCKETS: Connection from %s:%d.\n",
		  inet_ntoa(clientname.sin_addr), ntohs(clientname.sin_port));
	  if (FD_ISSET(i, &passwd_fd_set)) {
	    FD_SET(s, &active_fd_set);
	    FD_SET(s, &passwd_read_fd_set);
	  }
	  else {
	    FD_CLR(s, &new_fd_set);
	    if (i == master_socket) {
	      FD_SET(s, &active_fd_set);
	      FD_SET(s, &data_fd_set);
	      fprintf(stderr, "HANDLE_SOCKETS: New master from %s:%d.\n", 
		      inet_ntoa(clientname.sin_addr), ntohs(clientname.sin_port));
	      master_socket = s;
	    }
	    else if (FD_ISSET(i, &player_fd_set)) {
	      /* don't listen to the player any more in this loop */
	      FD_CLR(i, &player_fd_set);
	      player_change_socket(i, s);
	      player_change_status(s, STATUS_READY);
	    }
	  }
	}
	/* master / game data socket */
	else if (FD_ISSET(i, &data_fd_set)) {
	  /* master? */
	  if (master_active == ON && i == master_socket) {
	    size = recv(i, b, 1024, 0);
	    b[size] = 0;
	    if (master_command(b) == SUCCEEDED)
	      continue;
	    /* server exit */
	    return FAILED;
	  }
	  /* player? */
	  if (FD_ISSET(i, &player_fd_set)) {
	    size = recv(i, b, 1024, 0);
	    if (size <= 0)
	      continue;
	    b[size] = 0;

	    p = players;
	    while (p != NULL) {
	      if (p->socket == i)
		break;
	      p = p->next;
	    }
	    if (p == NULL)
	      return FAILED;

	    fprintf(stderr, "HANDLE_SOCKETS: Illegal message \"%s\" from player \"%s\".\n", b, p->go->name);
	    continue;
	  }
	}
	/* passwd query sockets */
	else if (FD_ISSET(i, &passwd_read_fd_set)) {
	  /* read the password and the world information */
	  size = recv(i, b, 1024, 0);
	  while (size < 4)
	    size += recv(i, b+size, 1024-size, 0);
	  read_int(b, &n);
	  /* get the whole message */
	  while (size < n)
	    size += recv(i, b+size, 1024-size, 0);

	  /* check that the world data matches */
	  for (size = 4; b[size] != 0; size++)
	    ;

	  size++;
	  n = 0;

	  /* how the world is generated (image / seed) */
	  if ((unsigned int)b[size++] != world_map)
	    n = 1;

	  /* the seed */
	  read_int(b+size, &x);
	  if ((unsigned int)x != world_seed)
	    n = 1;

	  /* player passwd? */
	  if (n == 0 && strcmp(session_passwd, b+4) == 0 && accept_clients == YES) {
	    while (1) {
	      generate_port(&po);
	      if (socket_create(&so, po, YES) == SUCCEEDED)
		break;
	    }

	    player_create(i, po, so);

	    /* login ok! */
	    n = 0;
	    write_int_a(b, &n, 4+1+4+strlen(world_name)+1);
	    b[n++] = 1;
	    write_int_a(b, &n, po);
	    sprintf(b+n, "%s%c", world_name, 0);
	    send(s, b, n+strlen(world_name)+1, 0);

	    FD_SET(so, &active_fd_set);
	    FD_SET(so, &player_fd_set);
	    FD_SET(so, &new_fd_set);
	  }
	  /* master passwd? */
	  else if (strcmp(master_passwd, b+4) == 0) {
	    if (master_active == ON) {
	      fprintf(stderr, "HANDLE_SOCKETS: Trying to use master passwd while the master is already logged in.\n");
	      send(i, "Error", 5, 0);
	    }
	    else {
	      while (1) {
		generate_port(&po);
		if (socket_create(&so, po, YES) == SUCCEEDED)
		  break;
	      }
	      master_port = po;
	      master_socket = so;
	      master_active = ON;
	      sprintf(b, "Welcome to port %d", po);
	      send(i, b, strlen(b), 0);
	      FD_SET(so, &active_fd_set);
	      FD_SET(so, &new_fd_set);
	    }
	  }
	  /* wrong passwd or wrong world data? */
	  else {
	    if (strcmp(session_passwd, b+4) == 0)
	      x = 0;
	    else
	      x = 1;

	    n = 0;
	    write_int_a(b, &n, 4+1+4+strlen(world_name)+1);
	    b[n++] = 0;
	    b[n++] = x;
	    b[n++] = world_map;
	    write_int_a(b, &n, world_seed);

	    send(s, b, n, 0);
	  }
	  close(i);
	  FD_CLR(i, &active_fd_set);
	  FD_CLR(i, &passwd_read_fd_set);
	}
      }
    }
  }

  return SUCCEEDED;
}


void signal_handler_int(int signum) {

  fprintf(stderr, "SIGNAL_HANDLER_INT: Shutdown\n");

  if (players != NULL || master_active == ON) {
    fprintf(stderr, "SIGNAL_HANDLER_INT: Broadcasting my death...\n");
    broadcast_death();
  }

  /* close the server */
  server_status = OFF;
}


int master_command(char *c) {

  int i;


  /* server exit? */
  if (c[0] == 'X') {
    fprintf(stderr, "MASTER_COMMAND: Close server.\n");
    broadcast_death();
    server_status = OFF;
    return SUCCEEDED;
  }
  /* terminal exit? */
  if (c[0] == 'x') {
    fprintf(stderr, "MASTER_COMMAND: Close terminal.\n");
    master_active = OFF;
    return SUCCEEDED;
  }
  /* close main port? */
  if (c[0] == 'C') {
    fprintf(stderr, "MASTER_COMMAND: Close main port %d.\n", main_port);
    FD_CLR(main_socket, &new_fd_set);
    FD_CLR(main_socket, &active_fd_set);
    FD_CLR(main_socket, &passwd_fd_set);
    socket_delete(main_socket);
    send(master_socket, "Y", 1, 0);
    main_port_active = OFF;
    return SUCCEEDED;
  }
  /* open main port? */
  if (c[0] == 'O') {
    fprintf(stderr, "MASTER_COMMAND: Open main port %d.\n", main_port);
    if (main_port_active == ON) {
      send(master_socket, "Y", 1, 0);
      return SUCCEEDED;
    }
    if (socket_create(&main_socket, main_port, YES) == FAILED) {
      send(master_socket, "N", 1, 0);
      return FAILED;
    }
    FD_SET(main_socket, &new_fd_set);
    FD_SET(main_socket, &active_fd_set);
    FD_SET(main_socket, &passwd_fd_set);
    main_port_active = ON;
    send(master_socket, "Y", 1, 0);
    return SUCCEEDED;
  }
  /* new main port? */
  if (c[0] == 'N') {
    sscanf(c + 1, "%d", &i);
    if (i < IPPORT_USERRESERVED || i > 0xFFFF) {
      fprintf(stderr, "MASTER_COMMAND: Suggested new main port %d is out of range.\n", i);
      send(master_socket, "N", 1, 0);
      return SUCCEEDED;
    }
    main_port = i;
    fprintf(stderr, "MASTER_COMMAND: New main port %d - ", main_port);
    if (socket_create(&i, main_port, YES) == FAILED) {
      send(master_socket, "N", 1, 0);
      fprintf(stderr, "FAILED\n");
      return FAILED;
    }
    send(master_socket, "Y", 1, 0);
    fprintf(stderr, "SUCCEEDED\n");
    if (main_port_active == ON) {
      FD_CLR(main_socket, &new_fd_set);
      FD_CLR(main_socket, &active_fd_set);
      FD_CLR(main_socket, &passwd_fd_set);
      socket_delete(main_socket);
    }
    main_socket = i;
    FD_SET(main_socket, &new_fd_set);
    FD_SET(main_socket, &active_fd_set);
    FD_SET(main_socket, &passwd_fd_set);
    return SUCCEEDED;
  }

  fprintf(stderr, "MASTER_COMMAND: Unknown command \"%s\".\n", c);

  return FAILED;
}


int broadcast_death(void) {

  struct player *p;
  char b[8];


  /* tell the players */
  p = players;
  while (p != NULL) {
    p->status = STATUS_QUIT;
    b[0] = 'D';
    player_write(p, b, 1);
    player_handle(p);
    p = p->next;
  }

  /* tell the master */
  if (master_active == ON) {
    send(master_socket, "D", 1, 0);
    recv(master_socket, b, 8, 0);
  }

  return SUCCEEDED;
}


int close_all_sockets(void) {

  struct player *p;


  p = players;
  while (p != NULL) {
    FD_CLR(p->socket, &player_fd_set);
    FD_CLR(p->socket, &active_fd_set);
    FD_CLR(p->socket, &data_fd_set);
    socket_delete(p->socket);
    p = p->next;
  }

  if (master_active == ON)
    socket_delete(master_socket);

  socket_delete(main_socket);

  return SUCCEEDED;
}


int world_init(void) {

  char c[256];
  FILE *f;
  int i;


  strcpy(c, "../data/wld/");
  strcat(c, world_name);
  strcat(c, "/info.txt");
  f = fopen(c, "rb");
  if (f == NULL) {
    fprintf(stderr, "WORLD_INIT: Could not open \"%s\".\n", c);
    return FAILED;
  }

  fscanf(f, "%d %d %d", &world_x, &world_y, &world_z);
  fclose(f);

  world_time = malloc(sizeof(unsigned int) * world_x*world_y*world_z);
  if (world_time == NULL) {
    fprintf(stderr, "WORLD_INIT: Out of memory error.\n");
    return FAILED;
  }

  for (i = 0; i < world_x*world_y*world_z; i++)
    world_time[i] = current_time;

  return SUCCEEDED;
}


void send_to_all(char *c, int i) {

  struct player *p;


  /* handle player i/o */
  p = players;
  while (p != NULL) {
    if (p->status == STATUS_READY)
      player_write(p, c, i);
    p = p->next;
  }
}


void send_to_all_but(struct player *l, char *c, int i) {

  struct player *p;


  /* handle player i/o */
  p = players;
  while (p != NULL) {
    if (p != l)
      if (p->status == STATUS_READY)
	player_write(p, c, i);
    p = p->next;
  }
}
