
/* (c) 2001 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkhscrollbar.h>
#include <gtk/gtkvscrollbar.h>
#include <gdk/gdkkeysyms.h>

#include "main.h"
#include "defines.h"


volatile sig_atomic_t alarm_status = ON;
volatile int connection_status = OFF;
struct itimerval timer_old, timer_new;
char master_passwd[256], server_name[256];
int server_port = 6789, client_port, actual_port, actual_socket;

GtkWidget *window_main = NULL, *main_vbox_1, *menubar_main, *entry_main;
GtkAccelGroup *accel_group_main;
GtkItemFactory *item_factory_main;
GtkItemFactoryEntry menu_items_main[] = {
  { "/_Main",              NULL,         NULL, 0, "<Branch>" },
  { "/Main/_Quit",         "<control>Q", quit_window_main_selected, 0, NULL }
};


int main(int argc, char *argv[]) {

  srand(time(NULL));

  if (signal(SIGINT, signal_handler_int) == SIG_IGN)
    return 1;
  if (signal(SIGALRM, signal_handler_alrm) == SIG_IGN)
    return 1;

  gtk_set_locale();

  timer_new.it_interval.tv_usec = 500000;
  timer_new.it_interval.tv_sec = 0;
  timer_new.it_value.tv_usec = 500000;
  timer_new.it_value.tv_sec = 0;
  setitimer(ITIMER_REAL, &timer_new, &timer_old);

  strcpy(master_passwd, "sugar!");
  strcpy(server_name, "localhost");

  connect_to_server();
  gtk_init(&argc, &argv);

  init_window_main();

  gtk_widget_show_all(window_main);

  gtk_main();

  gtk_exit(0);
  return 0;
}


int init_window_main(void) {

  /* main window */
  window_main = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  main_window_set_title();

  /* map panel */
  accel_group_main = gtk_accel_group_new();
  item_factory_main = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accel_group_main);
  gtk_item_factory_create_items(item_factory_main, sizeof(menu_items_main)/sizeof(menu_items_main[0]), menu_items_main, NULL);
  gtk_accel_group_attach(accel_group_main, GTK_OBJECT(window_main));
  menubar_main = gtk_item_factory_get_widget(item_factory_main, "<main>");

  /* entry */
  entry_main = gtk_entry_new();
  gtk_entry_set_editable(GTK_ENTRY(entry_main), TRUE);

  /* boxes */
  main_vbox_1 = gtk_vbox_new(FALSE, 0);

  /* construct window */
  gtk_container_add(GTK_CONTAINER(window_main), main_vbox_1);
  gtk_box_pack_start(GTK_BOX(main_vbox_1), menubar_main, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(main_vbox_1), entry_main, FALSE, FALSE, 0);

  /* signals */
  gtk_signal_connect(GTK_OBJECT(window_main), "delete_event", GTK_SIGNAL_FUNC(quit_window_main), NULL);
  gtk_signal_connect(GTK_OBJECT(entry_main), "activate", GTK_SIGNAL_FUNC(entry_enter_callback), NULL);

  return SUCCEEDED;
}


int connect_to_host(const char *hostname, int p, int s) {

  struct sockaddr_in name;
  struct hostent *hostinfo;


  name.sin_family = AF_INET;
  name.sin_port = htons(p);
  hostinfo = gethostbyname(hostname);
  if (hostinfo == NULL) {
      fprintf (stderr, "CONNECT_TO_HOST: Unknown host %s.\n", hostname);
      return FAILED;
  }
  name.sin_addr = *(struct in_addr *)hostinfo->h_addr;

  if (connect(s, (struct sockaddr *)&name, sizeof(name)) < 0) {
    fprintf(stderr, "CONNECT_TO_HOST: Connecting to %s:%d failed.\n", hostname, p);
    return FAILED;
  }

  fprintf(stderr, "CONNECT_TO_HOST: Connected to %s:%d.\n", hostname, p);

  return SUCCEEDED;
}


int create_new_socket(int *s, int p) {

  struct sockaddr_in name;


  *s = socket(PF_INET, SOCK_STREAM, 0);
  if (*s < 0) {
    fprintf(stderr, "CREATE_NEW_SOCKET: Error creating a new socket.\n");
    return FAILED;
  }

  name.sin_port = htons(p);
  name.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(*s, (struct sockaddr *)&name, sizeof(name)) < 0) {
    delete_socket(*s);
    fprintf(stderr, "CREATE_NEW_SOCKET: Error binding port %d to a new socket.\n", p);
    return FAILED;
  }

  fprintf(stderr, "CREATE_NEW_SOCKET: Obtained port %d.\n", p);

  return SUCCEEDED;
}


int delete_socket(int s) {

  if (shutdown(s, SHUT_RDWR) >= 0)
    return SUCCEEDED;
  else if (close(s) >= 0)
    return SUCCEEDED;

  return SUCCEEDED;
}


int generate_port(int *po) {

  static int c = 0;


  *po = rand() + c;
  *po = (*po % (0xFFFF - IPPORT_USERRESERVED)) + IPPORT_USERRESERVED;
  c++;

  return SUCCEEDED;
}


int parse_port(char *c, int *p) {

  while ((*c != 0) && ((*c < '0') || (*c > '9')))
    c++;

  if (*c == 0)
    return FAILED;

  *p = 0;
  while (*c != 0) {
    *p = *p * 10 + *c - '0';
    c++;
  }

  return SUCCEEDED;
}


int process_terminal_command(char *b) {

  char t[256];
  int i;


  sscanf(b, "%s", t);

  /* x - x - close terminal */
  if (strcmp(t, "x") == 0) {
    if (connection_status == ON)
      send(actual_socket, "x", 1, 0);
    fprintf(stderr, "READ_COMMANDLINE: Close terminal.\n");
    delete_socket(actual_socket);
    connection_status = OFF;
    gtk_main_quit();
    return SUCCEEDED;
  }
  /* cl - x - close connection */
  else if (strcmp(t, "cl") == 0) {
    if (connection_status == OFF) {
      fprintf(stderr, "READ_COMMANDLINE: Not connected.\n");
      return FAILED;
    }
    send(actual_socket, "x", 1, 0);
    fprintf(stderr, "READ_COMMANDLINE: Close connection.\n");
    delete_socket(actual_socket);
    connection_status = OFF;
    main_window_set_title();
    return SUCCEEDED;
  }
  /* co - handshake - connect to server */
  else if (strcmp(t, "co") == 0) {
    if (connection_status == ON) {
      fprintf(stderr, "READ_COMMANDLINE: Already connected to %s:%d.\n", server_name, actual_port);
      return FAILED;
    }
    i = connect_to_server();
    fprintf(stderr, "READ_COMMANDLINE: Connecting to server %s - ", server_name);
    if (i == SUCCEEDED) {
      fprintf(stderr, "SUCCEEDED.\n");
      return SUCCEEDED;
    }
    fprintf(stderr, "FAILED.\n");
    return FAILED;
  }
  /* cs - X - close server */
  else if (strcmp(t, "cs") == 0) {
    if (connection_status == OFF) {
      fprintf(stderr, "READ_COMMANDLINE: Not connected.\n");
      return FAILED;
    }
    send(actual_socket, "X", 1, 0);
    i = recv(actual_socket, b, 256, 0);
    b[i] = 0;
    fprintf(stderr, "READ_COMMANDLINE: Close server - ");
    if (b[0] == 'D') {
      send(actual_socket, "Y", 1, 0);
      fprintf(stderr, "SUCCEEDED.\n");
      connection_status = OFF;
      main_window_set_title();
      return SUCCEEDED;
    }
    send(actual_socket, "N", 1, 0);
    fprintf(stderr, "FAILED.\n");
    connection_status = ON;
    main_window_set_title();
    return FAILED;
  }
  /* cmp - C - close main port */
  else if (strcmp(t, "cmp") == 0) {
    if (connection_status == OFF) {
      fprintf(stderr, "READ_COMMANDLINE: Not connected.\n");
      return FAILED;
    }
    send(actual_socket, "C", 1, 0);
    i = recv(actual_socket, b, 256, 0);
    b[i] = 0;
    fprintf(stderr, "READ_COMMANDLINE: Close main port - ");
    if (b[0] == 'Y') {
      fprintf(stderr, "SUCCEEDED.\n");
      return SUCCEEDED;
    }
    fprintf(stderr, "FAILED.\n");
    return FAILED;
  }
  /* omp - O - open main port */
  else if (strcmp(t, "omp") == 0) {
    if (connection_status == OFF) {
      fprintf(stderr, "READ_COMMANDLINE: Not connected.\n");
      return FAILED;
    }
    send(actual_socket, "O", 1, 0);
    i = recv(actual_socket, b, 256, 0);
    b[i] = 0;
    fprintf(stderr, "READ_COMMANDLINE: Open main port - ");
    if (b[0] == 'Y') {
      fprintf(stderr, "SUCCEEDED.\n");
      return SUCCEEDED;
    }
    fprintf(stderr, "FAILED.\n");
    return FAILED;
  }
  /* nmp - N - new main port */
  else if (strcmp(t, "nmp") == 0) {
    sscanf(b, "%*s %d", &server_port);
    if (connection_status == ON) {
      sprintf(b, "N %d", server_port);
      send(actual_socket, b, strlen(b), 0);
      i = recv(actual_socket, b, 256, 0);
      b[i] = 0;
      fprintf(stderr, "READ_COMMANDLINE: New main port %d - ", server_port);
      if (b[0] == 'Y') {
	fprintf(stderr, "SUCCEEDED.\n");
	return SUCCEEDED;
      }
      fprintf(stderr, "FAILED.\n");
      return FAILED;
    }
    else
      fprintf(stderr, "READ_COMMANDLINE: New main port %d (locally).\n", server_port);
    return SUCCEEDED;
  }
  /* csn - [none] - change server name */
  else if (strcmp(t, "csn") == 0) {
    sscanf(b, "%*s %s", t);
    if (connection_status == ON) {
      fprintf(stderr, "READ_COMMANDLINE: Already connected to %s:%d.\n", server_name, actual_port);
      return FAILED;
    }
    strcpy(server_name, t);
    fprintf(stderr, "READ_COMMANDLINE: Changed server name to \"%s\".\n", server_name);
    return SUCCEEDED;
  }
  /* chp - [none] - change passwd */
  else if (strcmp(t, "chp") == 0) {
    sscanf(b, "%*s %s", t);
    if (connection_status == ON) {
      fprintf(stderr, "READ_COMMANDLINE: Already connected to %s:%d.\n", server_name, actual_port);
      return FAILED;
    }
    strcpy(master_passwd, t);
    fprintf(stderr, "READ_COMMANDLINE: New master passwd \"%s\".\n", master_passwd);
    return SUCCEEDED;
  }
  /* unknown command */
  else {
    fprintf(stderr, "READ_COMMANDLINE: Unknown command. Available commands:\n");
    fprintf(stderr, "                  cl, co, cs, cmp, chp, omp, nmp, nsn, x\n");
    return FAILED;
  }

  return FAILED;
}


int connect_to_server(void) {

  char b[256];
  int s, size;


  connection_status = OFF;

  while (1) {
    generate_port(&client_port);
    if (create_new_socket(&s, client_port) == SUCCEEDED)
      break;
  }

  if (connect_to_host(server_name, server_port, s) == FAILED) {
    main_window_set_title();
    return FAILED;
  }

  send(s, master_passwd, strlen(master_passwd), 0);
  size = recv(s, b, 256, 0);
  b[size] = 0;

  if (strcmp(b, "No bonus") == 0) {
    fprintf(stderr, "CONNECT_TO_SERVER: Invalid master passwd \"%s\".\n", master_passwd);
    delete_socket(s);
    main_window_set_title();
    return FAILED;
  }
  else if (strcmp(b, "Error") == 0) {
    fprintf(stderr, "CONNECT_TO_SERVER: Can't connect as master.\n");
    delete_socket(s);
    main_window_set_title();
    return FAILED;
  }

  delete_socket(s);

  parse_port(b, &actual_port);

  while (1) {
    generate_port(&client_port);
    if (create_new_socket(&actual_socket, client_port) == SUCCEEDED)
      break;
  }

  if (connect_to_host(server_name, actual_port, actual_socket) == FAILED) {
    main_window_set_title();
    return FAILED;
  }

  connection_status = ON;
  main_window_set_title();

  return SUCCEEDED;
}


void signal_handler_int(int signum) {

  if (connection_status == ON) {
    send(actual_socket, "x", 1, 0);
    delete_socket(actual_socket);
    connection_status = OFF;
  }
}


void signal_handler_alrm(int signum) {

  struct timeval t;
  fd_set read_fd_set;
  char b[256];
  int i;


  if (signal(SIGALRM, signal_handler_alrm) == SIG_IGN) {
    fprintf(stderr, "SIGNAL_HANDLE_ALRM: Error re-establishing the alarm handler.\n");
    return;
  }

  if (alarm_status == OFF)
    return;

  if (connection_status == ON) {
    t.tv_sec = 0;
    t.tv_usec = 0;
    FD_ZERO(&read_fd_set);
    FD_SET(actual_socket, &read_fd_set);
    i = select(FD_SETSIZE, &read_fd_set, NULL, NULL, &t);
    if (i <= 0)
      return;
    i = recv(actual_socket, b, 256, 0);
    b[i] = 0;
    if (b[0] == 'D') {
      printf("SIGNAL_HANDLER_ALRM: Server died.\n");
      send(actual_socket, "Y", 1, 0);
      delete_socket(actual_socket);
      connection_status = OFF;
      main_window_set_title();
    }
  }
}


void quit_window_main_selected(GtkWidget *widget, gpointer data) {

  quit_window_main(widget, data);
}


gint quit_window_main(GtkWidget *widget, gpointer data) {

  if (connection_status == ON) {
    send(actual_socket, "x", 1, 0);
    delete_socket(actual_socket);
    connection_status = OFF;
  }

  gtk_main_quit();

  return TRUE;
}


int main_window_set_title(void) {

  char t[256];


  if (window_main == NULL)
    return SUCCEEDED;

  if (connection_status == OFF)
    gtk_window_set_title(GTK_WINDOW(window_main), "Terminal - Not connected");
  else {
    sprintf(t, "Terminal - %s:%d", server_name, actual_port);
    gtk_window_set_title(GTK_WINDOW(window_main), t);
  }

  return SUCCEEDED;
}


gint entry_enter_callback(GtkWidget *widget, gpointer data) {

  char b[256];


  strcpy(b, gtk_entry_get_text(GTK_ENTRY(widget)));
  gtk_entry_set_text(GTK_ENTRY(widget), "");
  process_terminal_command(b);

  return TRUE;
}
