
/* (c) 2001-2003 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>

#include <SDL/SDL.h>
#include <SDL/SDL_video.h>
#include <SDL/SDL_keysym.h>
#include <GL/gl.h>

#include "defines.h"
#include "gui.h"


extern int nmeshes;

const SDL_VideoInfo *sdl_video_info = NULL;
int status = STATUS_OK, view = 0, view_mode = VIEW_MODE_FILLED;

double player_x = 0, player_y = 0;


int init_window_main(void) {

  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0) {
    fprintf(stderr, "INIT_WINDOW_MAIN: Video initialization failed: %s\n", SDL_GetError());
    return FAILED;
  }

  sdl_video_info = SDL_GetVideoInfo();

  SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

  if (SDL_SetVideoMode(SCREEN_X, SCREEN_Y, sdl_video_info->vfmt->BitsPerPixel, SDL_OPENGL | SDL_FULLSCREEN) == 0) {
    fprintf(stderr, "INIT_WINDOW_MAIN: Video mode initialization failed: %s\n", SDL_GetError());
    return FAILED;
  }

  SDL_EnableUNICODE(1);

  return SUCCEEDED;
}


int keyboard_io(void) {

  SDL_Event event;
  int i;


  while (SDL_PollEvent(&event)) {
    switch (event.type) {

    case SDL_KEYDOWN:
      i = event.key.keysym.sym;
      switch (i) {
      case SDLK_w:
	status |= STATUS_DISTANCE_MINUS;
	break;
      case SDLK_s:
	status |= STATUS_DISTANCE_PLUS;
	break;
      case SDLK_a:
	if (view > 0)
	  view--;
	break;
      case SDLK_d:
	if (view < nmeshes-1)
	  view++;
	break;
      case SDLK_q:
	if (view_mode == VIEW_MODE_LINES)
	  view_mode = VIEW_MODE_FILLED;
	else
	  view_mode = VIEW_MODE_LINES;
	break;
      case SDLK_ESCAPE:
	status |= STATUS_QUIT;
	break;
      }
      break;

    case SDL_KEYUP:
      i = event.key.keysym.sym;
      switch (i) {
      case SDLK_w:
	status &= ~STATUS_DISTANCE_MINUS;
	break;
      case SDLK_s:
	status &= ~STATUS_DISTANCE_PLUS;
	break;
      }
      break;
    }
  }

  return SUCCEEDED;
}


int sdl_filter_events(const SDL_Event *event) {

  if (event->type == SDL_QUIT)
    return 1;

  if (event->type == SDL_MOUSEMOTION) {
    player_x += event->motion.xrel / (double)SCREEN_X * 360;
    player_y += event->motion.yrel / (double)SCREEN_Y * 360;
    if (player_x < 0)
      player_x += 360;
    if (player_x >= 360)
      player_x -= 360;
    if (player_y < 0)
      player_y += 360;
    if (player_y >= 360)
      player_y += 360;
  }

  return 1;
}
