
/* (c) 2001-2002 ville helin <vhelin@cc.hut.fi> */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <GL/gl.h>
#include <GL/glext.h>

#include "defines.h"
#include "../common/defines.h"
#include "../common/boxen.h"
#include "txt.h"



static int txt_id = 0, txt_mem = 0;

struct txt_extra_info *txt_extra_info_list = NULL;
struct txt_object *txt_first, *txt_last, **txt_table;
char txt_base_dir[256];
int txt_id_conversion = ON, dirty_textures = YES;



int txt_init(char *base) {

  char n[256], t[256];
  struct txt_extra_info *r;
  float d, e, a;
  FILE *f;
  int i, c;


  strcpy(txt_base_dir, base);
  sprintf(n, "%sextrainfo.txt%c", base, 0);

  txt_table = malloc(sizeof(struct txt_object *) * MAX_TEXTURES);
  if (txt_table == NULL) {
    fprintf(stderr, "TXT_INIT: Out of memory error.\n");
    return FAILED;
  }

  /* empty texture table */
  for (i = 0; i < MAX_TEXTURES; i++)
    txt_table[i] = NULL;

  /* load the texture assigned friction values */
  f = fopen(n, "rb");
  if (f == NULL) {
    fprintf(stderr, "TXT_INIT: Could not open file \"%s\".\n", n);
    return FAILED;
  }

  /* get pass the comment section */
  fscanf(f, "%s", t);
  while (strcmp(t, "GO!") != 0)
    if (fscanf(f, "%s", t) == EOF)
      break;

  if (strcmp(t, "GO!") != 0) {
    fprintf(stderr, "TXT_INIT: Could not find marker \"GO!\" inside the file \"%s\".\n", n);
    fclose(f);
    return FAILED;
  }

  /* read the data rows */
  while (fscanf(f, "%s %f %f %f %d", t, &d, &e, &a, &c) != EOF) {
    r = malloc(sizeof(struct txt_extra_info));
    if (r == NULL) {
      fprintf(stderr, "TXT_INIT: Out of memory error.\n");
      fclose(f);
      return FAILED;
    }
    strcpy(r->name, t);
    r->friction = d;
    r->speed = e;
    r->alpha = a;
    r->collision = c;
    r->next = txt_extra_info_list;
    txt_extra_info_list = r;
  }

  fclose(f);

  return SUCCEEDED;
}


int txt_create(char *n, int *id, int x, int y) {

  struct txt_object *t;
  int i;


  /* no name -> it must be a new one! */
  if (n != NULL) {
    for (i = 0; i < MAX_TEXTURES; i++) {
      t = txt_table[i];
      if (t != NULL && strcmp(n, t->name) == 0) {
	*id = i;
	return SUCCEEDED;
      }
    }
  }

  t = malloc(sizeof(struct txt_object));
  if (t == NULL) {
    fprintf(stderr, "TXT_CREATE: Out of memory error.\n");
    return FAILED;
  }

  t->txt = malloc(x*y*4);
  if (t->txt == NULL) {
    fprintf(stderr, "TXT_CREATE: Out of memory error.\n");
    free(t);
    return FAILED;
  }

  if (txt_get_next_free_id(&t->id) == FAILED) {
    free(t);
    return FAILED;
  }

  t->x = x;
  t->y = y;
  t->dirty = NO;
  t->speed = 1;
  t->friction = 0;
  t->alpha = 1;
  t->ogl_name = 0;
  *id = t->id;

  if (n != NULL)
    strcpy(t->name, n);
  else
    strcpy(t->name, "noname");

  txt_table[t->id] = t;

  /* collect statistics */
  txt_mem += x*y*4;

  return SUCCEEDED;
}


int txt_free_txt(int id) {

  struct txt_object *t;


  if (txt_table[id] == NULL) {
    fprintf(stderr, "TXT_FREE_ID: No previous texture %d.\n", id);
    return FAILED;
  }

  t = txt_table[id];
  if (t->ogl_name != 0)
    glDeleteTextures(1, &(t->ogl_name));
  /* collect statistics */
  txt_mem -= t->x*t->y*4;
  free(t);
  txt_table[id] = NULL;

  return SUCCEEDED;
}


int txt_get_next_free_id(int *id) {

  int i, x;


  i = txt_id;
  for (x = 0; x < MAX_TEXTURES; x++) {
    if (txt_table[i] == NULL) {
      *id = i;
      txt_id = (i+1) % MAX_TEXTURES;
      return SUCCEEDED;
    }
    i = (i+1) % MAX_TEXTURES;
  }

  /* no more free id numbers */
  fprintf(stderr, "TXT_GET_NEXT_FREE_ID: Out of texture ID numbers.\n");

  return FAILED;
}


int txt_load(char *n, int *id) {

  struct txt_extra_info *f;
  struct txt_object *t;
  char c[256];
  int i, p, div;


  for (i = 0; i < MAX_TEXTURES; i++) {
    t = txt_table[i];
    if (t != NULL && strcmp(n, t->name) == 0) {
      *id = i;
      return SUCCEEDED;
    }
  }

  t = malloc(sizeof(struct txt_object));
  if (t == NULL) {
    fprintf(stderr, "TXT_LOAD: Out of memory.\n");
    return FAILED;
  }

  /* localize extra information */
  f = txt_extra_info_list;
  while (f != NULL) {
    if (strcmp(f->name, n) == 0)
      break;
    f = f->next;
  }

  if (f == NULL) {
    fprintf(stderr, "TXT_LOAD: No extra information for texture \"%s\".\n", n);
    t->friction = 0;
    t->speed = 1;
    t->alpha = 1;
    t->status = TXT_STATUS_NORMAL;
  }
  else {
    t->friction = f->friction;
    t->speed = f->speed;
    if (f->alpha == 1.0)
      t->status = TXT_STATUS_NORMAL;
    else
      t->status = TXT_STATUS_ALPHA;
    if (f->collision == 0)
      t->status |= TXT_STATUS_NO_COLLISION;
  }

  t->dirty = NO;
  if (txt_get_next_free_id(&t->id) == FAILED) {
    free(f);
    return FAILED;
  }

  *id = t->id;
  strcpy(t->name, n);
  strcpy(c, txt_base_dir);
  strcat(c, n);

  boxen_read(c, (unsigned char **)&t->txt, &t->x, &t->y, &t->size, &t->type, &t->mipmaps);

  glGenTextures(1, &t->ogl_name);
  if (t->ogl_name == 0) {
    fprintf(stderr, "TXT_LOAD: Error creating texture \"%s\".\n", t->name);
    free(t);
    return FAILED;
  }

  glBindTexture(GL_TEXTURE_2D, t->ogl_name);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

  if (t->mipmaps > 1) {
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, t->mipmaps-1);
  }
  else
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  for (p = 0, i = 0, div = 1; i < t->mipmaps; i++) {
    glCompressedTexImage2DARB(GL_TEXTURE_2D, i, t->type, t->x/div, t->y/div, 0, t->size[i], t->txt + p);
    p += t->size[i];
    div *= 2;
  }

  txt_table[t->id] = t;

  return SUCCEEDED;
}


int txt_get(int id, struct txt_object **t) {

  *t = txt_table[id];
  if (*t == NULL)
    fprintf(stderr, "TXT_GET: Unknown texture %d.\n", id);

  return SUCCEEDED;
}


int txt_free_all(void) {

  struct txt_object *t;
  int i;


  fprintf(stderr, "TXT_FREE_ALL: Freeing %dkB worth of textures.\n", txt_mem/1024);

  for (i = 0; i < MAX_TEXTURES; i++) {
    t = txt_table[i];
    if (t != NULL)
      txt_free_txt(i);
  }

  return SUCCEEDED;
}


int txt_refresh(void) {

  struct txt_object *t;
  int i;


  if (dirty_textures == NO)
    return SUCCEEDED;

  for (i = 0; i < MAX_TEXTURES; i++) {
    t = txt_table[i];
    if (t == NULL)
      continue;

    if (t->dirty == YES) {
      if (t->ogl_name == 0) {
	glGenTextures(1, &t->ogl_name);
	if (t->ogl_name == 0) {
	  fprintf(stderr, "TXT_REFRESH: Error creating texture \"%s\".\n", t->name);
	  return FAILED;
	}
      }
      glBindTexture(GL_TEXTURE_2D, t->ogl_name);
      glTexImage2D(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGBA_ARB, t->x, t->y, 0, GL_RGBA, GL_UNSIGNED_BYTE, t->txt);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      t->dirty = NO;
    }
  }

  dirty_textures = NO;

  return SUCCEEDED;
}
